from django.utils.translation import ugettext as _
#bots-modules
import botslib
from botsconfig import *

def grammarread(editype,grammarname,typeofgrammarfile='grammars'):
    ''' dispatch function for class Grammar and subclasses.
        read whole grammar or only syntax (parameter 'typeofgrammarfile').
        directory from where grammar is via parameter 'typeofgrammarfile'.
    '''
    try:
        classtocall = globals()[editype]
    except KeyError:
        raise botslib.GrammarError(_(u'Read grammar for editype "%(editype)s" messagetype "%(messagetype)s", but editype is unknown.'),
                                        {'editype':editype,'messagetype':grammarname})
    if typeofgrammarfile == 'grammars':
        # read grammar for messagetype first, than syntax for envelope.
        messagegrammar = classtocall(typeofgrammarfile='grammars',editype=editype,grammarname=grammarname)
        #Get right syntax: start with classtocall.defaultsyntax, update with envelope.syntax, update again with messagetype.syntax
        syntax = classtocall.defaultsyntax.copy()
        envelope = messagegrammar.syntax.get('envelope') or classtocall.defaultsyntax['envelope']
        if envelope and envelope != grammarname:
            try:
                envelopegrammar = classtocall(typeofgrammarfile='grammars',editype=editype,grammarname=envelope)
                syntax.update(envelopegrammar.syntax)
            except:
                pass
        syntax.update(messagegrammar.syntax)
        messagegrammar.syntax = syntax
        return messagegrammar
    elif typeofgrammarfile == 'envelope':
        #used when reading grammar for enveloping/outgoing. For 'noenvelope' this will not be done.
        # read syntax for messagetype first, than grammar for envelope.
        messagegrammar = classtocall(typeofgrammarfile='grammars',editype=editype,grammarname=grammarname)
        #Get right syntax: start with classtocall.defaultsyntax, update with envelope.syntax, update again with messagetype.syntax
        syntax = classtocall.defaultsyntax.copy()
        envelope = messagegrammar.syntax.get('envelope') or classtocall.defaultsyntax['envelope']
        if envelope != grammarname:
            try:
                envelopegrammar = classtocall(typeofgrammarfile='grammars',editype=editype,grammarname=envelope)
                syntax.update(envelopegrammar.syntax)
            except:
                envelopegrammar = messagegrammar
        syntax.update(messagegrammar.syntax)
        envelopegrammar.syntax = syntax
        return envelopegrammar
    else:   #typeofgrammarfile == 'partners':
        return classtocall(typeofgrammarfile='partners',editype=editype,grammarname=grammarname)


class Grammar(object):
    ''' Class for translation grammar. A grammar contains the description of an edi-file; this is used in reading or writing an edi file.
        Description of the grammar file: see user manual.
        The grammar is read from a grammar file.
        A grammar file has several grammar parts , eg 'structure' and 'recorddefs'.
        Grammar parts is either in the grammar part itself or a imported from another grammar-file (eg the edifact segments .

        in a grammar 'structure' is a list of dicts describing the sequence and relationships between the record(group)s:
            attributes of each record(group) in structure:
            -   ID       record id
            -   MIN      min #occurences record or group
            -   MAX      max #occurences record of group
            -   LEVEL    child-records
            added after reading the grammar (so: not in grammar-file):
            -   MPATH    mpath of record
            -   FIELDS   (added from recordsdefs via lookup)
        in a grammar 'recorddefs' describes the (sub) fields for the records:
        -   'recorddefs' is a dict where key is the recordID, value is list of (sub) fields
            each (sub)field is a tuple of (field or subfield)
            field is tuple of (ID, MANDATORY, LENGTH, FORMAT)
            subfield is tuple of (ID, MANDATORY, tuple of fields)

        every grammar-file is read once (default python import-machinery).
        The information in a grammar is checked and manipulated by bots.
        if a structure or recorddef has already been read, Bots skips most of the checks.
    '''
    def __init__(self,typeofgrammarfile,editype,grammarname):
        self.module,self.grammarname = botslib.botsimport(typeofgrammarfile,editype,grammarname)
        #get syntax from grammar file
        syntaxfromgrammar = getattr(self.module, 'syntax',{})
        if not isinstance(syntaxfromgrammar,dict):
            raise botslib.GrammarError(_(u'Grammar "%(grammar)s": syntax is not a dict{}.'),
                                        {'grammar':self.grammarname})
        self.syntax = syntaxfromgrammar.copy()  #copy to get independent syntax-object

        if typeofgrammarfile == 'partners':
            return
        #init rest of grammar
        self.nextmessage = getattr(self.module, 'nextmessage',None)
        self.nextmessage2 = getattr(self.module, 'nextmessage2',None)
        self.nextmessageblock = getattr(self.module, 'nextmessageblock',None)
        if self.nextmessage is None:
            if self.nextmessage2 is not None:
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s": if nextmessage2: nextmessage has to be used.'),
                                            {'grammar':self.grammarname})
        else:
            if self.nextmessageblock is not None:
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s": nextmessageblock and nextmessage not both allowed.'),
                                            {'grammar':self.grammarname})
        if self._get_fromsyntax_or_defaultsyntax('has_structure'):
            try:
                self._dorecorddefs()
            except:
                self.recorddefs['BOTS_1$@#%_error'] = True                  #mark structure has been read with errors
                raise
            else:
                self.recorddefs['BOTS_1$@#%_error'] = False                 #mark structure has been read and checked
            try:
                self._dostructure()
            except AttributeError:  #if grammarpart does not exist set to None; test required grammarpart elsewhere
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s": no structure, is required.'),
                                            {'grammar':self.grammarname})
            except:
                self.structure[0]['error'] = True                #mark the structure as having errors
                raise
        self.extrachecks()

    def _dorecorddefs(self):
        ''' 1. check the recorddefinitions for validity.
            2. adapt in field-records: normalise length lists, set bool ISFIELD, etc
        '''
        try:
            self.recorddefs = getattr(self.module, 'recorddefs')
        except AttributeError:
            raise botslib.GrammarError(_(u'Grammar "%(grammar)s": no recorddefs.'),
                                        {'grammar':self.grammarname})
        if not isinstance(self.recorddefs,dict):
            raise botslib.GrammarError(_(u'Grammar "%(grammar)s": recorddefs is not a dict{}.'),
                                            {'grammar':self.grammarname})
        #check if grammar is read & checked earlier in this run. If so, we can skip all checks.
        if 'BOTS_1$@#%_error' in self.recorddefs:   #if checked before
            if self.recorddefs['BOTS_1$@#%_error']:     #if grammar had errors
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s" has error that is already reported in this run.'),
                                            {'grammar':self.grammarname})
            return      #no error, skip checks
        for recordid ,fields in self.recorddefs.iteritems():
            if not isinstance(recordid,basestring):
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s": is not a string.'),
                                                {'grammar':self.grammarname,'record':recordid})
            if not recordid:
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s": recordid with empty string.'),
                                            {'grammar':self.grammarname,'record':recordid})
            if not isinstance(fields,list):
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s": no correct fields found.'),
                                            {'grammar':self.grammarname,'record':recordid})
            if isinstance(self,(xml,json)):
                if len (fields) < 1:
                    raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s": too few fields.'),
                                                {'grammar':self.grammarname,'record':recordid})
            else:
                if len (fields) < 2:
                    raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s": too few fields.'),
                                                {'grammar':self.grammarname,'record':recordid})

            has_botsid = False   #to check if BOTSID is present
            fieldnamelist = []  #to check for double fieldnames
            for field in fields:
                self._checkfield(field,recordid)
                if not field[ISFIELD]:  # if composite
                    for sfield in field[SUBFIELDS]:
                        self._checkfield(sfield,recordid)
                        if sfield[ID] in fieldnamelist:
                            raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s": field "%(field)s" appears twice. Field names should be unique within a record.'),
                                                        {'grammar':self.grammarname,'record':recordid,'field':sfield[ID]})
                        fieldnamelist.append(sfield[ID])
                else:
                    if field[ID] == 'BOTSID':
                        has_botsid = True
                    if field[ID] in fieldnamelist:
                        raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s": field "%(field)s" appears twice. Field names should be unique within a record.'),
                                                        {'grammar':self.grammarname,'record':recordid,'field':field[ID]})
                    fieldnamelist.append(field[ID])

            if not has_botsid:   #there is no field 'BOTSID' in record
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s": no field BOTSID.'),
                                                {'grammar':self.grammarname,'record':recordid})

    def _checkfield(self,field,recordid):
        #'normalise' field: make list equal length
        len_field = len(field)
        if len_field == 3:  # that is: composite
            field += [None,False,None,None,'A',1]
        elif len_field == 4:               # that is: field (not a composite)
            field += [True,0,0,'A',1]
        #each field is now equal length list
        elif len_field == 9:               # this happens when there are errors in a table and table is read again
            raise botslib.GrammarError(_(u'Grammar "%(grammar)s": error in grammar; error is already reported in this run.'),
                                            {'grammar':self.grammarname})
        else:
            raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s": list has invalid number of arguments.'),
                                            {'grammar':self.grammarname,'record':recordid,'field':field[ID]})
        if not isinstance(field[ID],basestring) or not field[ID]:
            raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s": fieldID has to be a string.'),
                                            {'grammar':self.grammarname,'record':recordid,'field':field[ID]})
        if isinstance(field[MANDATORY],basestring):
            if field[MANDATORY] not in 'MC':
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s": mandatory/conditional must be "M" or "C".'),
                                                {'grammar':self.grammarname,'record':recordid,'field':field[ID]})
            field[MANDATORY] = 0 if field[MANDATORY]=='C' else 1
        elif isinstance(field[MANDATORY],tuple):
            if not isinstance(field[MANDATORY][0],basestring):
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s": mandatory/conditional must be "M" or "C".'),
                                                {'grammar':self.grammarname,'record':recordid,'field':field[ID]})
            if field[MANDATORY][0] not in 'MC':
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s": mandatory/conditional must be "M" or "C".'),
                                                {'grammar':self.grammarname,'record':recordid,'field':field[ID]})
            if not isinstance(field[MANDATORY][1],int):
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s": number of repeats must be integer.'),
                                                {'grammar':self.grammarname,'record':recordid,'field':field[ID]})
            field[MAXREPEAT] = field[MANDATORY][1]
            field[MANDATORY] = 0 if field[MANDATORY][0] == 'C' else 1
        else:
            raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s": mandatory/conditional has to be a string (or tuple in case of repeating field).'),
                                            {'grammar':self.grammarname,'record':recordid,'field':field[ID]})
        if field[ISFIELD]:  # that is: field, and not a composite
            #get MINLENGTH (from tuple or if fixed
            if isinstance(field[LENGTH],(int,float)):
                if isinstance(self,fixed):
                    field[MINLENGTH] = field[LENGTH]
            elif isinstance(field[LENGTH],tuple):
                if not isinstance(field[LENGTH][0],(int,float)):
                    raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s": min length "%(min)s" has to be a number.'),
                                                    {'grammar':self.grammarname,'record':recordid,'field':field[ID],'min':field[LENGTH][0]})
                if not isinstance(field[LENGTH][1],(int,float)):
                    raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s": max length "%(max)s" has to be a number.'),
                                                    {'grammar':self.grammarname,'record':recordid,'field':field[ID],'max':field[LENGTH][1]})
                if field[LENGTH][0] > field[LENGTH][1]:
                    raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s": min length "%(min)s" must be > max length "%(max)s".'),
                                                    {'grammar':self.grammarname,'record':recordid,'field':field[ID],'min':field[LENGTH][0],'max':field[LENGTH][1]})
                field[MINLENGTH] = field[LENGTH][0]
                field[LENGTH] = field[LENGTH][1]
            else:
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s": length "%(len)s" has to be number or (min,max).'),
                                                {'grammar':self.grammarname,'record':recordid,'field':field[ID],'len':field[LENGTH]})
            if field[LENGTH] < 1:
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s": length "%(len)s" has to be at least 1.'),
                                                {'grammar':self.grammarname,'record':recordid,'field':field[ID],'len':field[LENGTH]})
            if field[MINLENGTH] < 0:
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s": minlength "%(len)s" has to be at least 0.'),
                                                {'grammar':self.grammarname,'record':recordid,'field':field[ID],'len':field[LENGTH]})
            #format
            if not isinstance(field[FORMAT],basestring):
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s": format "%(format)s" has to be a string.'),
                                                {'grammar':self.grammarname,'record':recordid,'field':field[ID],'format':field[FORMAT]})
            self._manipulatefieldformat(field,recordid)
            if field[BFORMAT] in 'NIR':
                if isinstance(field[LENGTH],float):
                    length,nrdecimals = divmod(field[LENGTH],1)     #divide by 1 to get whole number and leftover
                    field[DECIMALS] = int( round(nrdecimals*10) )   #fill DECIMALS with leftover*10. Does not work for more than 9 decimal places...
                    field[LENGTH] = int(length)
                    if field[DECIMALS] >= field[LENGTH]:
                        raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s": field length "%(len)s" has to be greater that nr of decimals "%(decimals)s".'),
                                                        {'grammar':self.grammarname,'record':recordid,'field':field[ID],'decimals':field[DECIMALS]})
                if isinstance(field[MINLENGTH],float):
                    field[MINLENGTH] = int(field[MINLENGTH]//1)
            else:   #if format 'R', A, D, T
                if isinstance(field[LENGTH],float):
                    raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s": if format "%(format)s", no length "%(len)s".'),
                                                    {'grammar':self.grammarname,'record':recordid,'field':field[ID],'format':field[FORMAT],'len':field[LENGTH]})
                if isinstance(field[MINLENGTH],float):
                    raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s": if format "%(format)s", no minlength "%(len)s".'),
                                                    {'grammar':self.grammarname,'record':recordid,'field':field[ID],'format':field[FORMAT],'len':field[MINLENGTH]})
        else:       #check composite
            if not isinstance(field[SUBFIELDS],list):
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s": is a composite field, has to have subfields.'),
                                                {'grammar':self.grammarname,'record':recordid,'field':field[ID]})
            if len(field[SUBFIELDS]) < 2:
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s" has < 2 sfields.'),
                                                {'grammar':self.grammarname,'record':recordid,'field':field[ID]})

    def _linkrecorddefs2structure(self,structure):
        ''' recursive
            for each record in structure: add the pointer to the right recorddefinition.
        '''
        for i in structure:
            try:
                i[FIELDS] = self.recorddefs[i[ID]]
            except KeyError:
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s": in recorddef no record "%(record)s".'),
                                            {'grammar':self.grammarname,'record':i[ID]})
            if LEVEL in i:
                self._linkrecorddefs2structure(i[LEVEL])

    def _dostructure(self):
        ''' 1. check the structure for validity.
            2. adapt in structure: Add keys: mpath, count
            3. remember that structure is checked and adapted (so when grammar is read again, no checking/adapt needed)
        '''
        self.structure = getattr(self.module, 'structure')
        if len(self.structure) != 1:                        #every structure has only 1 root!!
            raise botslib.GrammarError(_(u'Grammar "%(grammar)s", in structure: only one root record allowed.'),{'grammar':self.grammarname})
        #check if structure is read & checked earlier in this run. If so, we can skip all checks.
        if 'error' in self.structure[0]:
            pass        # grammar has been read before, but there are errors. Do nothing here, same errors will be raised again.
        elif MPATH in self.structure[0]:
            return      # grammar has been read before, with no errors. Do no checks.
        self._checkstructure(self.structure,[])
        if self._get_fromsyntax_or_defaultsyntax('checkcollision'):
            self._checkbackcollision(self.structure)
            self._checknestedcollision(self.structure)
        self._checkbotscollision(self.structure)
        self._linkrecorddefs2structure(self.structure)

    def _checkstructure(self,structure,mpath):
        ''' Recursive
            1.   Check structure.
            2.   Add keys: mpath, count
        '''
        if not isinstance(structure,list):
            raise botslib.GrammarError(_(u'Grammar "%(grammar)s", in structure, at "%(mpath)s": not a list.'),
                                        {'grammar':self.grammarname,'mpath':mpath})
        for i in structure:
            if not isinstance(i,dict):
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", in structure, at "%(mpath)s": record should be a dict: "%(record)s".'),
                                            {'grammar':self.grammarname,'mpath':mpath,'record':i})
            if ID not in i:
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", in structure, at "%(mpath)s": record without ID: "%(record)s".'),
                                            {'grammar':self.grammarname,'mpath':mpath,'record':i})
            if not isinstance(i[ID],basestring):
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", in structure, at "%(mpath)s": recordid of record is not a string: "%(record)s".'),
                                            {'grammar':self.grammarname,'mpath':mpath,'record':i})
            if not i[ID]:
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", in structure, at "%(mpath)s": recordid of record is empty: "%(record)s".'),
                                            {'grammar':self.grammarname,'mpath':mpath,'record':i})
            if MIN not in i:
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", in structure, at "%(mpath)s": record without MIN: "%(record)s".'),
                                            {'grammar':self.grammarname,'mpath':mpath,'record':i})
            if MAX not in i:
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", in structure, at "%(mpath)s": record without MAX: "%(record)s".'),
                                            {'grammar':self.grammarname,'mpath':mpath,'record':i})
            if not isinstance(i[MIN],int):
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", in structure, at "%(mpath)s": record where MIN is not whole number: "%(record)s".'),
                                            {'grammar':self.grammarname,'mpath':mpath,'record':i})
            if not isinstance(i[MAX],int):
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", in structure, at "%(mpath)s": record where MAX is not whole number: "%(record)s".'),
                                            {'grammar':self.grammarname,'mpath':mpath,'record':i})
            if not i[MAX]:
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", in structure, at "%(mpath)s": MAX is zero: "%(record)s".'),
                                            {'grammar':self.grammarname,'mpath':mpath,'record':i})
            if i[MIN] > i[MAX]:
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", in structure, at "%(mpath)s": record where MIN > MAX: "%(record)s".'),
                                            {'grammar':self.grammarname,'mpath':mpath,'record':unicode(i)[:100]})
            i[MPATH] = mpath + [i[ID]]
            if LEVEL in i:
                self._checkstructure(i[LEVEL],i[MPATH])

    def _checkbackcollision(self,structure,collision=None):
        ''' Recursive.
            Check if grammar has back-collision problem. A message with collision problems is ambiguous.
            Case 1:  AAA BBB AAA
            Case 2:  AAA     BBB
                     BBB CCC
        '''
        if not collision:
            collision = []
        headerissave = False
        for i in structure:
            #~ print 'check back segment:',i[MPATH], 'with list:',collision
            if i[ID] in collision:
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", in structure: back-collision detected at record "%(mpath)s".'),
                                            {'grammar':self.grammarname,'mpath':i[MPATH]})
            if i[MIN]:
                headerissave = True
                if i[MIN] == i[MAX]:    #so: fixed number of occurences; can not lead to collision as  is always clear where in structure record is
                    collision = []      #NOTE: this is mainly used for MIN=1, MAX=1
                else:
                    collision = [i[ID]] #previous records do not cause collision.
            else:
                collision.append(i[ID])
            if LEVEL in i:
                if i[MIN] == i[MAX] == 1:
                    returncollision,returnheaderissave = self._checkbackcollision(i[LEVEL])
                else:
                    returncollision,returnheaderissave = self._checkbackcollision(i[LEVEL],[i[ID]])
                collision.extend(returncollision)
                if returnheaderissave and i[ID] in collision:  #if one of segment(groups) is required, there is always a segment after the header segment; so remove header from nowcollision:
                    collision.remove(i[ID])
        return collision,headerissave    #collision is used to update on higher level; cleared indicates the header segment can not collide anymore

    def _checkbotscollision(self,structure):
        ''' Recursive.
            Within one level: if twice the same tag: use BOTSIDNR.
        '''
        collision = {}
        for i in structure:
            if i[ID] in collision:
                #~ raise botslib.GrammarError(_(u'Grammar "%(grammar)s", in structure: bots-collision detected at record "%(mpath)s".'),{'grammar':self.grammarname,'mpath':i[MPATH]})
                i[BOTSIDNR] = unicode(collision[i[ID]] + 1)
                collision[i[ID]] = collision[i[ID]] + 1
            else:
                i[BOTSIDNR] = u'1'
                collision[i[ID]] = 1
            if LEVEL in i:
                self._checkbotscollision(i[LEVEL])
        return

    def _checknestedcollision(self,structure,collision=None):
        ''' Recursive.
            Check if grammar has nested-collision problem. A message with collision problems is ambiguous.
            Case 1: AAA
                    BBB CCC
                        AAA
        '''
        if not collision:
            levelcollision = []
        else:
            levelcollision = collision[:]
        for i in reversed(structure):
            if LEVEL in i:
                if i[MIN] == i[MAX] == 1:
                    isa_safeheadersegment = self._checknestedcollision(i[LEVEL],levelcollision)
                else:
                    isa_safeheadersegment = self._checknestedcollision(i[LEVEL],levelcollision + [i[ID]])
            else:
                isa_safeheadersegment = False
            #~ print 'check nested segment',checkthissegment, i[MPATH], 'with list',levelcollision
            if isa_safeheadersegment or i[MIN] == i[MAX]:    #fixed number of occurences. this can be handled umambigiously: no check needed
                pass   #no check needed
            elif i[ID] in levelcollision:
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s", in structure: nesting collision detected at record "%(mpath)s".'),
                                            {'grammar':self.grammarname,'mpath':i[MPATH]})
            if i[MIN]:
                levelcollision = []   #empty uppercollision
        return not bool(levelcollision)

    def extrachecks(self):
        ''' default function, some subclasses have the actual checks.'''
        pass

    def display(self,structure,level=0):
        ''' Draw grammar, with indentation for levels.
            For debugging.
        '''
        for i in structure:
            print 'Record: ',i[MPATH],i
            for field in i[FIELDS]:
                print '    Field: ',field
            if LEVEL in i:
                self.display(i[LEVEL],level+1)

    #bots interpretats the format from the grammer; left side are the allowed values; right side are the internal forams bots uses.
    #the list directly below are the default values for the formats, subclasses can have their own list.
    #this makes it possible to use x12-formats for x12, edifact-formats for edifact etc
    formatconvert = {
        'A':'A',        #alfanumerical
        'AN':'A',       #alfanumerical
        #~ 'AR':'A',       #right aligned alfanumerical field, used in fixed records.
        'D':'D',        #date
        'DT':'D',       #date-time
        'T':'T',        #time
        'TM':'T',       #time
        'N':'N',        #numerical, fixed decimal. Fixed nr of decimals; if no decimal used: whole number, integer
        #~ 'NL':'N',       #numerical, fixed decimal. In fixed format: no preceding zeros, left aligned,
        #~ 'NR':'N',       #numerical, fixed decimal. In fixed format: preceding blancs, right aligned,
        'R':'R',        #numerical, any number of decimals; the decimal point is 'floating'
        #~ 'RL':'R',       #numerical, any number of decimals. fixed: no preceding zeros, left aligned
        #~ 'RR':'R',       #numerical, any number of decimals. fixed: preceding blancs, right aligned
        'I':'I',        #numercial, implicit decimal
        }
    def _manipulatefieldformat(self,field,recordid):
        try:
            field[BFORMAT] = self.formatconvert[field[FORMAT]]
        except KeyError:
            raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record "%(record)s", field "%(field)s": format "%(format)s" has to be one of "%(keys)s".'),
                                        {'grammar':self.grammarname,'record':recordid,'field':field[ID],'format':field[FORMAT],'keys':self.formatconvert.keys()})
                                        
    def _get_fromsyntax_or_defaultsyntax(self,value):
        return self.syntax.get(value) or self.__class__.defaultsyntax.get(value)

#grammar subclasses. contain the defaultsyntax
class test(Grammar):
    ''' For unit tests '''
    defaultsyntax = {
        'has_structure':True,   #is True, read structure, recorddef, check these
        'checkcollision':True,
        'noBOTSID':False,
        }
class csv(Grammar):
    def extracheck(self):
        if self._get_fromsyntax_or_defaultsyntax('noBOTSID') and len(self.recorddefs) != 1:
            raise botslib.GrammarError(_(u'Grammar "%(grammar)s": if syntax["noBOTSID"]: there can be only one record in recorddefs.'),
                                            {'grammar':self.grammarname})
        if self.nextmessageblock is not None and len(self.recorddefs) != 1:
            raise botslib.GrammarError(_(u'Grammar "%(grammar)s": if nextmessageblock: there can be only one record in recorddefs.'),
                                            {'grammar':self.grammarname})
    defaultsyntax = {
        'add_crlfafterrecord_sep':'',
        'allow_lastrecordnotclosedproperly':False,  #in csv sometimes the last record is no closed correctly. This is related to communciation over email. Beware: when using this, other checks will not be enforced!
        'charset':'utf-8',
        'checkcharsetin':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'checkcharsetout':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'contenttype':'text/csv',
        'decimaal':'.',
        'envelope':'',
        'escape':"",
        'field_sep':':',
        'forcequote': 1,            #(if quote_char is set) 0:no force: only quote if necessary:1:always force: 2:quote if alfanumeric
        'merge':True,
        'noBOTSID':False,           #allow csv records without record ID.
        'pass_all':True,            #(csv only) if only one recordtype and no nextmessageblock: would pass record for record to mapping. this fixes that.
        'quote_char':"'",
        'record_sep':"\r\n",        #better is  "\n" (got some strange errors for this?)
        'skip_char':'',
        'skip_firstline':False,     #often first line in CSV is fieldnames. Usage: either False/True, or number of lines. If True, number of lines is 1
        'triad':'',
        'wrap_length':0,     #for producing wrapped format, where a file consists of fixed length records ending with crr/lf. Often seen in mainframe, as400
        #settings needed as defaults, but not useful for this editype 
        'checkunknownentities': True,
        'record_tag_sep':"",    #Tradacoms/GTDI
        'reserve':'',
        'sfield_sep':'',
        #bots internal, never change/overwrite
        'has_structure':True,   #is True, read structure, recorddef, check these
        'checkcollision':True,
        'lengthnumericbare':False,
        'stripfield_sep':False,
        }
class excel(csv):
    pass
class fixed(Grammar):
    def extracheck(self):
        if self._get_fromsyntax_or_defaultsyntax('noBOTSID') and len(self.recorddefs) != 1:
            raise botslib.GrammarError(_(u'Grammar "%(grammar)s": if syntax["noBOTSID"]: there can be only one record in recorddefs.'),
                                            {'grammar':self.grammarname})
        if self.nextmessageblock is not None and len(self.recorddefs) != 1:
            raise botslib.GrammarError(_(u'Grammar "%(grammar)s": if nextmessageblock: there can be only one record in recorddefs.'),
                                            {'grammar':self.grammarname})
    formatconvert = {
        'A':'A',        #alfanumerical
        'AN':'A',       #alfanumerical
        'AR':'A',       #right aligned alfanumerical field, used in fixed records.
        'D':'D',        #date
        'DT':'D',       #date-time
        'T':'T',        #time
        'TM':'T',       #time
        'N':'N',        #numerical, fixed decimal. Fixed nr of decimals; if no decimal used: whole number, integer
        'NL':'N',       #numerical, fixed decimal. In fixed format: no preceding zeros, left aligned,
        'NR':'N',       #numerical, fixed decimal. In fixed format: preceding blancs, right aligned,
        'R':'R',        #numerical, any number of decimals; the decimal point is 'floating'
        'RL':'R',       #numerical, any number of decimals. fixed: no preceding zeros, left aligned
        'RR':'R',       #numerical, any number of decimals. fixed: preceding blancs, right aligned
        'I':'I',        #numercial, implicit decimal
        }
    defaultsyntax = {
        'charset':'us-ascii',
        'checkcharsetin':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'checkcharsetout':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'checkfixedrecordtoolong':True,
        'checkfixedrecordtooshort':False,
        'contenttype':'text/plain',
        'decimaal':'.',
        'envelope':'',
        'merge':True,
        'noBOTSID':False,           #allow fixed records without record ID.
        'triad':'',
        #settings needed as defaults, but not useful for this editype 
        'add_crlfafterrecord_sep':'',
        'checkunknownentities': True,
        'escape':'',
        'field_sep':'',
        'forcequote':0,         #csv only
        'quote_char':"",
        'record_sep':"\r\n",
        'record_tag_sep':"",    #Tradacoms/GTDI
        'reserve':'',
        'sfield_sep':'',
        'skip_char':'',
        #bots internal, never change/overwrite
        'has_structure':True,   #is True, read structure, recorddef, check these
        'checkcollision':True,
        'lengthnumericbare':False,
        'stripfield_sep':False,
        }
    is_first_record = True
    def _linkrecorddefs2structure(self,structure):
        ''' for class fixed: more checking etc is done.
            recursive
            for each record in structure: add the pointer to the right recorddefinition.
        '''
        for i in structure:
            try:
                i[FIELDS] = self.recorddefs[i[ID]]
            except KeyError:
                raise botslib.GrammarError(_(u'Grammar "%(grammar)s": in recorddef no record "%(record)s".'),
                                            {'grammar':self.grammarname,'record':i[ID]})
            #determine start/end of BOTSID; check if pos and length BOTSID is the same for all records.
            position_in_record = 0
            for field in i[FIELDS]:
                if field[ID] == 'BOTSID':
                    if self.is_first_record:     #set startrecordID, endrecordID
                        self.is_first_record = False
                        startrecordID = position_in_record
                        endrecordID = position_in_record + field[LENGTH]
                        self.syntax['startrecordID'] = startrecordID            #also change in the copy made.
                        self.syntax['endrecordID'] = endrecordID
                        #original grammar syntax has to be changed
                        if not hasattr(self.module, 'syntax'):      
                            self.module.syntax = {}
                        self.module.syntax['startrecordID'] = startrecordID
                        self.module.syntax['endrecordID'] = endrecordID
                        #~ syntaxfromgrammar = getattr(self.module, 'syntax',{})      #original grammar syntax has to be changed, so get it.
                    else:        #check startrecordID, endrecordID
                        if self.syntax['startrecordID'] != position_in_record or self.syntax['endrecordID'] != position_in_record + field[LENGTH]:
                            raise botslib.GrammarError(_(u'Grammar "%(grammar)s", record %(key)s: position and length of BOTSID should be equal in all records.'),
                                                    {'grammar':self.grammarname,'key':i[ID]})
                    break
                position_in_record += field[LENGTH]
            #calculate recordlength
            i[F_LENGTH] = sum([field[LENGTH] for field in i[FIELDS]])
            #go recursive
            if LEVEL in i:
                self._linkrecorddefs2structure(i[LEVEL])
class idoc(fixed):
    defaultsyntax = {
        'automaticcount':True,
        'charset':'us-ascii',
        'checkcharsetin':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'checkcharsetout':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'checkfixedrecordtoolong':False,
        'checkfixedrecordtooshort':False,
        'contenttype':'text/plain',
        'decimaal':'.',
        'envelope':'',
        'merge':False,
        'MANDT':'0',
        'DOCNUM':'0',
        #settings needed as defaults, but not useful for this editype 
        'add_crlfafterrecord_sep':'',
        'checkunknownentities': True,
        'escape':'',
        'field_sep':'',
        'forcequote':0,         #csv only
        'noBOTSID':False,           #allow fixed records without record ID.
        'quote_char':"",
        'record_sep':"\r\n",
        'record_tag_sep':"",    #Tradacoms/GTDI
        'reserve':'',
        'sfield_sep':'',
        'skip_char':'',
        'triad':'',
        #bots internal, never change/overwrite
        'has_structure':True,   #is True, read structure, recorddef, check these
        'checkcollision':True,
        'lengthnumericbare':False,
        'stripfield_sep':False,
        }
class xml(Grammar):
    def extracheck(self):
        if not self._get_fromsyntax_or_defaultsyntax('envelope') and self._get_fromsyntax_or_defaultsyntax('merge'):
            raise botslib.GrammarError(_(u'Grammar "%(grammar)s": in this xml grammar merge is "True" but no (user) enveloping is specified. This will lead to invalid xml files'),
                                            {'grammar':self.grammarname})
    defaultsyntax = {
        'attributemarker':'__',
        'charset':'utf-8',
        'checkcharsetin':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'checkcharsetout':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'checkunknownentities': True,
        'contenttype':'text/xml ',
        'decimaal':'.',
        'DOCTYPE':'',                   #doctype declaration to use in xml header. 'DOCTYPE': 'mydoctype SYSTEM "mydoctype.dtd"'  will lead to: <!DOCTYPE mydoctype SYSTEM "mydoctype.dtd">
        'envelope':'',
        'extra_character_entity':{},    #additional character entities to resolve when parsing XML; mostly html character entities. Example: {'euro':u'','nbsp':unichr(160),'apos':u'\u0027'}
        'indented':False,               #False: xml output is one string (no cr/lf); True: xml output is indented/human readable
        'merge':False,
        'namespace_prefixes':None,  #to over-ride default namespace prefixes (ns0, ns1 etc) for outgoing xml. is a list, consisting of tuples, each tuple consists of prefix and uri.
                                    #Example: 'namespace_prefixes':[('orders','http://www.company.com/EDIOrders'),]
        'processing_instructions': None,    #to generate processing instruction in xml prolog. is a list, consisting of tuples, each tuple consists of type of instruction and text for instruction.
                                            #Example: 'processing_instructions': [('xml-stylesheet' ,'href="mystylesheet.xsl" type="text/xml"'),('type-of-ppi' ,'attr1="value1" attr2="value2"')]
                                            #leads to this output in xml-file:  <?xml-stylesheet href="mystylesheet.xsl" type="text/xml"?><?type-of-ppi attr1="value1" attr2="value2"?>
        'standalone':None,      #as used in xml prolog; values: 'yes' , 'no' or None (not used)
        'triad':'',
        'version':'1.0',        #as used in xml prolog
        #settings needed as defaults, but not useful for this editype 
        'add_crlfafterrecord_sep':'',
        'escape':'',
        'field_sep':'',
        'forcequote':0,                 #csv only
        'quote_char':"",
        'record_sep':"",
        'record_tag_sep':"",    #Tradacoms/GTDI
        'reserve':'',
        'sfield_sep':'',
        'skip_char':'',
        #bots internal, never change/overwrite
        'has_structure':True,   #is True, read structure, recorddef, check these
        'checkcollision':False,
        'lengthnumericbare':False,
        'stripfield_sep':False,
        }
class xmlnocheck(xml):
    defaultsyntax = {
        'attributemarker':'__',
        'charset':'utf-8',
        'checkcharsetin':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'checkcharsetout':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'checkunknownentities': False,
        'contenttype':'text/xml ',
        'decimaal':'.',
        'DOCTYPE':'',                   #doctype declaration to use in xml header. DOCTYPE = 'mydoctype SYSTEM "mydoctype.dtd"'  will lead to: <!DOCTYPE mydoctype SYSTEM "mydoctype.dtd">
        'envelope':'',
        'extra_character_entity':{},    #additional character entities to resolve when parsing XML; mostly html character entities. Example: {'euro':u'','nbsp':unichr(160),'apos':u'\u0027'}
        'indented':False,               #False: xml output is one string (no cr/lf); True: xml output is indented/human readable
        'merge':False,
        'namespace_prefixes':None,  #to over-ride default namespace prefixes (ns0, ns1 etc) for outgoing xml. is a list, consisting of tuples, each tuple consists of prefix and uri.
                                    #Example: 'namespace_prefixes':[('orders','http://www.company.com/EDIOrders'),]
        'processing_instructions': None,    #to generate processing instruction in xml prolog. is a list, consisting of tuples, each tuple consists of type of instruction and text for instruction.
                                            #Example: processing_instructions': [('xml-stylesheet' ,'href="mystylesheet.xsl" type="text/xml"'),('type-of-ppi' ,'attr1="value1" attr2="value2"')]
                                            #leads to this output in xml-file:  <?xml-stylesheet href="mystylesheet.xsl" type="text/xml"?><?type-of-ppi attr1="value1" attr2="value2"?>
        'standalone':None,      #as used in xml prolog; values: 'yes' , 'no' or None (not used)
        'triad':'',
        'version':'1.0',        #as used in xml prolog
        #settings needed as defaults, but not useful for this editype 
        'add_crlfafterrecord_sep':'',
        'escape':'',
        'field_sep':'',
        'forcequote':0,                 #csv only
        'quote_char':"",
        'record_sep':"",
        'record_tag_sep':"",    #Tradacoms/GTDI
        'reserve':'',
        'sfield_sep':'',
        'skip_char':'',
        #bots internal, never change/overwrite
        'has_structure':False,   #is True, read structure, recorddef, check these
        'checkcollision':False,
        'lengthnumericbare':False,
        'stripfield_sep':False,
        }
class templatehtml(Grammar):
    defaultsyntax = {
        'charset':'utf-8',
        'checkcharsetin':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'checkcharsetout':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'contenttype':'text/xml',
        'decimaal':'.',
        'envelope':'templatehtml',
        'envelope-template':'',
        'merge':True,
        #settings needed as defaults, but not useful for this editype 
        'add_crlfafterrecord_sep':'',
        'checkunknownentities': True,
        'escape':'',
        'field_sep':'',
        'forcequote':0, #csv only
        'quote_char':"",
        'print_as_row':[],  #to indicate what should be printed as a table with 1 row per record (instead of 1 record->1 table)
        'record_sep':"",
        'record_tag_sep':"",    #Tradacoms/GTDI
        'reserve':'',
        'sfield_sep':'',
        'skip_char':'',
        'triad':'',
        #bots internal, never change/overwrite
        'has_structure':False,   #is True, read structure, recorddef, check these
        'checkcollision':False,
        'lengthnumericbare':False,
        'stripfield_sep':False,
        }
class edifact(Grammar):
    defaultsyntax = {
        'add_crlfafterrecord_sep':'\r\n',
        'charset':'UNOA',
        'checkcharsetin':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'checkcharsetout':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'contenttype':'application/EDIFACT',
        'decimaal':'.',
        'envelope':'edifact',
        'escape':'?',
        'field_sep':'+',
        'forceUNA' : False,
        'merge':True,
        'record_sep':"'",
        'reserve':'*',
        'sfield_sep':':',
        'skip_char':'\r\n',
        'version':'3',
        'UNB.S001.0080':'',
        'UNB.S001.0133':'',
        'UNB.S002.0007':'14',
        'UNB.S002.0008':'',
        'UNB.S002.0042':'',
        'UNB.S003.0007':'14',
        'UNB.S003.0014':'',
        'UNB.S003.0046':'',
        'UNB.S005.0022':'',
        'UNB.S005.0025':'',
        'UNB.0026':'',
        'UNB.0029':'',
        'UNB.0031':'',
        'UNB.0032':'',
        'UNB.0035':'0',
        #settings needed as defaults, but not useful for this editype 
        'checkunknownentities': True,
        'forcequote':0, #csv only
        'quote_char':'',
        'record_tag_sep':"",    #Tradacoms/GTDI
        'triad':'',
        #bots internal, never change/overwrite
        'has_structure':True,   #is True, read structure, recorddef, check these
        'checkcollision':True,
        'lengthnumericbare':True,
        'stripfield_sep':True,
        }
    formatconvert = {
        'A':'A',
        'AN':'A',
        'N':'R',
        }
class x12(Grammar):
    defaultsyntax = {
        'add_crlfafterrecord_sep':'\r\n',
        'charset':'us-ascii',
        'checkcharsetin':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'checkcharsetout':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'contenttype':'application/X12',
        'decimaal':'.',
        'envelope':'x12',
        'escape':'',
        'field_sep':'*',
        'functionalgroup'    :  'XX',
        'merge':True,
        'record_sep':"~",
        'replacechar':'',       #if separator found, replace by this character; if replacechar is an empty string: raise error
        'reserve':'^',
        'sfield_sep':'>',    #advised '\'?
        'skip_char':'\r\n',
        'version':'00403',
        'ISA01':'00',
        'ISA02':'          ',
        'ISA03':'00',
        'ISA04':'          ',
        'ISA05':'01',
        'ISA07':'01',
        'ISA11':'U',        #since ISA version 00403 this is the reserve/repetition separator. Bots does not use this anymore for ISA version >00403
        'ISA14':'1',
        'ISA15':'P',
        'GS07':'X',
        #settings needed as defaults, but not useful for this editype 
        'checkunknownentities': True,
        'forcequote':0, #csv only
        'quote_char':'',
        'record_tag_sep':"",    #Tradacoms/GTDI
        'triad':'',
        #bots internal, never change/overwrite
        'has_structure':True,   #is True, read structure, recorddef, check these
        'checkcollision':True,
        'lengthnumericbare':True,
        'stripfield_sep':True,
        }
    formatconvert = {
        'AN':'A',
        'DT':'D',
        'TM':'T',
        'N':'I',
        'N0':'I',
        'N1':'I',
        'N2':'I',
        'N3':'I',
        'N4':'I',
        'N5':'I',
        'N6':'I',
        'N7':'I',
        'N8':'I',
        'N9':'I',
        'R':'R',
        'B':'A',
        'ID':'A',
        }
    def _manipulatefieldformat(self,field,recordid):
        super(x12,self)._manipulatefieldformat(field,recordid)
        if field[BFORMAT] == 'I':
            if field[FORMAT][1:]:
                field[DECIMALS] = int(field[FORMAT][1])
            else:
                field[DECIMALS] = 0
class json(Grammar):
    defaultsyntax = {
        'charset':'utf-8',
        'checkcharsetin':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'checkcharsetout':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'checkunknownentities': True,
        'contenttype':'application/json',
        'decimaal':'.',
        'defaultBOTSIDroot':'ROOT',
        'envelope':'',
        'indented':False,               #False:  output is one string (no cr/lf); True:  output is indented/human readable
        'merge':False,
        'triad':'',
        #settings needed as defaults, but not useful for this editype 
        'add_crlfafterrecord_sep':'',
        'escape':'',
        'field_sep':'',
        'forcequote':0, #csv only
        'quote_char':"",
        'record_sep':"",
        'record_tag_sep':"",    #Tradacoms/GTDI
        'reserve':'',
        'sfield_sep':'',
        'skip_char':'',
        #bots internal, never change/overwrite
        'has_structure':True,   #is True, read structure, recorddef, check these
        'checkcollision':False,
        'lengthnumericbare':False,
        'stripfield_sep':False,
        }
class jsonnocheck(json):
    defaultsyntax = {
        'charset':'utf-8',
        'checkcharsetin':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'checkcharsetout':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'checkunknownentities': False,
        'contenttype':'application/json',
        'decimaal':'.',
        'defaultBOTSIDroot':'ROOT',
        'envelope':'',
        'indented':False,               #False:  output is one string (no cr/lf); True: output is indented/human readable
        'merge':False,
        'triad':'',
        #settings needed as defaults, but not useful for this editype 
        'add_crlfafterrecord_sep':'',
        'escape':'',
        'field_sep':'',
        'forcequote':0, #csv only
        'quote_char':"",
        'record_sep':"",
        'record_tag_sep':"",    #Tradacoms/GTDI
        'reserve':'',
        'sfield_sep':'',
        'skip_char':'',
        #bots internal, never change/overwrite
        'has_structure':False,   #is True, read structure, recorddef, check these
        'checkcollision':False,
        'lengthnumericbare':False,
        'stripfield_sep':False,
       }
class tradacoms(Grammar):
    defaultsyntax = {
        'add_crlfafterrecord_sep':'\r\n',
        'charset':'us-ascii',
        'checkcharsetin':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'checkcharsetout':'strict', #strict, ignore or botsreplace (replace with char as set in bots.ini).
        'contenttype':'application/text',
        'decimaal':'.',
        'envelope':'tradacoms',
        'escape':'?',
        'field_sep':'+',
        'merge':False,
        'record_sep':"'",
        'record_tag_sep':"=",    #Tradacoms/GTDI
        'sfield_sep':':',
        'STX.STDS1':'ANA',
        'STX.STDS2':'1',
        'STX.FROM.02':'',
        'STX.UNTO.02':'',
        'STX.APRF':'',
        'STX.PRCD':'',
        #settings needed as defaults, but not useful for this editype 
        'checkunknownentities': True,
        'forcequote':0, #csv only
        'indented':False,               #False:  output is one string (no cr/lf); True:  output is indented/human readable
        'quote_char':'',
        'reserve':'',
        'skip_char':'\r\n',
        'triad':'',
        #bots internal, never change/overwrite
        'has_structure':True,   #is True, read structure, recorddef, check these
        'checkcollision':True,
        'lengthnumericbare':True,
        'stripfield_sep':True,
        }
    formatconvert = {
        'X':'A',
        '9':'R',
        '9V9':'I',
        }



#!/usr/bin/env python
''' Start bots-engine2: do not use database for logging and configuration. (so: no GUI).
    Parameters are hard-coded for now (inpath, infilename, outpath, outfilename, editype, messagetype)
    Translation information (as from translation table: mapping script, outgoing editype etc) is eithr hard-coded now, or via translate table.
'''

import sys
import os
import atexit
import logging
from django.utils.translation import ugettext as _
#bots-modules
import botslib
import botsglobal
import botsinit
import warnings

def abspathdata(filename):
    ''' abspathdata if filename incl dir: return absolute path; else (only filename): return absolute path (datadir).
        for engine2 the current abspathdata is overwritten, as this uses subdirectories.
    '''
    if '/' in filename: #if filename already contains path
        return botslib.join(filename)
    else:
        return botslib.join(data_storage,filename)
botslib.abspathdata = abspathdata

def start():
    ''' sysexit codes:
        0: OK, no errors
        1: (system) errors incl parsing of command line arguments
        2: bots ran OK, but there are errors/process errors  in the run
        3: Database is locked, but "maxruntime" has not been exceeded.
    '''
    #NOTE: bots directory should always be on PYTHONPATH - otherwise it will not start.
    #********command line arguments**************************
    usage = '''
    This is "%(name)s" version %(version)s, part of Bots open source edi translator (http://bots.sourceforge.net).
    Does the actual translations and communications; it's the workhorse. It does not have a fancy interface.

    Usage:
        %(name)s  [config-option]
    Config-option:
        -c<directory>        directory for configuration files (default: config).

    '''%{'name':os.path.basename(sys.argv[0]),'version':botsglobal.version}
    configdir = 'config'
    for arg in sys.argv[1:]:
        if arg.startswith('-c'):
            configdir = arg[2:]
            if not configdir:
                print 'Error: configuration directory indicated, but no directory name.'
                sys.exit(1)
        elif arg in ["?", "/?",'-h', '--help'] or arg.startswith('-'):
            print usage
            sys.exit(0)
    #***********end handling command line arguments**************************
    
    botsinit.generalinit(configdir)     #find locating of bots, configfiles, init paths etc.
    #set working directory to bots installation. advantage: when using relative paths it is clear that this point paths within bots installation. 
    os.chdir(botsglobal.ini.get('directories','botspath'))

    #**************initialise logging******************************
    process_name = 'engine2'
    botsglobal.logger = botsinit.initenginelogging(process_name)
    atexit.register(logging.shutdown)
    for key,value in botslib.botsinfo():    #log info about environement, versions, etc
        botsglobal.logger.info(u'%(key)s: "%(value)s".',{'key':key,'value':value})

    #**************connect to database**********************************
    try:
        botsinit.connect()
    except Exception as msg:
        botsglobal.logger.exception(_(u'Could not connect to database. Database settings are in bots/config/settings.py. Error: "%(msg)s".'),{'msg':msg})
        sys.exit(1)
    else:
        botsglobal.logger.info(_(u'Connected to database.'))
        atexit.register(botsglobal.db.close)

    warnings.simplefilter('error', UnicodeWarning)
        
    #import global scripts for bots-engine
    try:
        userscript,scriptname = botslib.botsimport('routescripts','botsengine')
    except botslib.BotsImportError:      #userscript is not there; other errors like syntax errors are not catched
        userscript = scriptname = None
    #***acceptance tests: initialiase acceptance user script******************************
    acceptance_userscript = acceptance_scriptname = None
    if botsglobal.ini.getboolean('acceptance','runacceptancetest',False):
        botsglobal.logger.info(_(u'This run is an acceptance test - as indicated in option "runacceptancetest" in bots.ini.'))
        try:
            acceptance_userscript,acceptance_scriptname = botslib.botsimport('routescripts','bots_acceptancetest')
        except botslib.BotsImportError:
            botsglobal.logger.info(_(u'In acceptance test there is no script file "bots_acceptancetest.py" to check the results of the acceptance test.'))

    try:
        #~ botslib.prepare_confirmrules()
        #in acceptance tests: run a user script before running eg to clean output directories******************************
        botslib.tryrunscript(acceptance_userscript,acceptance_scriptname,'pretest')
        botslib.tryrunscript(userscript,scriptname,'pre')
        errorinrun = engine2_run()
    except Exception as msg:
        botsglobal.logger.exception(_(u'Severe error in bots system:\n%(msg)s'),{'msg':unicode(msg)})    #of course this 'should' not happen.
        sys.exit(1)
    else:
        if errorinrun:
            sys.exit(2) #indicate: error(s) in run(s)
        else:
            sys.exit(0) #OK



import glob
import shutil
import datetime
try:
    import cElementTree as ET
except ImportError:
    try:
        import elementtree.ElementTree as ET
    except ImportError:
        try:
            from xml.etree import cElementTree as ET
        except ImportError:
            from xml.etree import ElementTree as ET
import inmessage
import outmessage
import transform
import envelope
from botsconfig import *

data_storage = 'botssys/data2'


def engine2_run():
    #~ botsglobal.ini.set('directories','data',botslib.join(data_storage))
    print datetime.datetime.now()
    botslib.dirshouldbethere(data_storage)
    run = get_control_information()
    read_incoming(run)
    translate(run)
    mergemessages(run)
    write_outgoing(run)
    trace(run)
    report(run)
    cleanup(run)
    print datetime.datetime.now()
    return run.errorinrun

class Run(object):
    inpath = None
    outpath = None
    incoming = []  #as received
    translated = [] #result of translations
    outgoing = []  #as enveloped & outgoing
    errorinrun = 0

def get_control_information():
    ''' information from:
        - command line parameters
        - parsing xml file
        - via http: via url?
        For the moment: hard-coded
    '''
    run = Run()
    run.inpath = 'botssys/infile/edifact_xml/edifact'
    run.infilename = '*'
    run.outpath = 'botssys/outfile'
    run.outfilename = '{messagetype}_{infile:name}_{datetime:%Y%m%d}_*.{editype}'
    run.translation = dict(editype='edifact',messagetype='edifact')     #no tscript etc: will do lookup in translate-table
    #~ run.translation = dict(editype='edifact',messagetype='edifact',tscript='orders_edifact2xml',toeditype='xml',tomessagetype='orders')
    return run

def read_incoming(run):
    outputdir = botslib.join(run.inpath,run.infilename)
    filelist = [filename for filename in glob.iglob(outputdir) if os.path.isfile(filename)]
    filelist.sort()
    for infilename in filelist:
        try:
            filename = transform.unique('bots_file_name')
            abs_filename = botslib.abspathdata(filename)
            shutil.copy(infilename,abs_filename)          #move if to be delted
        except:
            txt = botslib.txtexc()
        else:
            txt = ''    #no errors
        finally:
            run.incoming.append({'infilename':infilename,'filename':filename,'error':txt,'editype':run.translation['editype'],'messagetype':run.translation['messagetype']})
                
def translate(run):
    for messagedict in run.incoming:
        try:
            #read whole edi-file: read, parse and made into a inmessage-object. Message is represented as a tree (inmessage.root is the root of the tree).
            edifile = inmessage.parse_edi_file(frompartner='',
                                                topartner='',
                                                filename=messagedict['filename'],
                                                messagetype=run.translation['messagetype'],
                                                testindicator='',
                                                editype=run.translation['editype'],
                                                charset='',
                                                alt='',
                                                fromchannel='',
                                                idroute='',
                                                command='')
            edifile.checkforerrorlist() #no exception if infile has been lexed and parsed OK else raises an error

            #~ if int(routedict['translateind']) == 3: #parse & passthrough; file is parsed, partners are known, no mapping, does confirm. 
                #~ raise botslib.GotoException('dummy')
            #~ continue    #file is parsed; no errors. no translation
            
            #edifile.ta_info contains info about incoming file: QUERIES, charset etc
            for inn_splitup in edifile.nextmessage():   #splitup messages in parsed edifile
                try:
                    #inn_splitup.ta_info: parameters from inmessage.parse_edi_file(), syntax-information and parse-information
                    number_of_loops_with_same_alt = 0
                    while 1:    #continue as long as there are (alt-)translations
                        #lookup the translation************************
                        tscript,toeditype,tomessagetype = 'orders_edifact2xml' ,'xml','orders'
                        if 'tscript' in run.translation:
                            tscript = run.translation['tscript']
                            toeditype = run.translation['toeditype']
                            tomessagetype = run.translation['tomessagetype']
                        else:
                            tscript,toeditype,tomessagetype = botslib.lookup_translation(fromeditype=inn_splitup.ta_info['editype'],
                                                                                frommessagetype=inn_splitup.ta_info['messagetype'],
                                                                                frompartner=inn_splitup.ta_info['frompartner'],
                                                                                topartner=inn_splitup.ta_info['topartner'],
                                                                                alt=inn_splitup.ta_info['alt'])
                            
                        #run mapping script************************
                        filename_translated = transform.unique('bots_file_name')
                        out_translated = outmessage.outmessage_init(editype=toeditype,
                                                                    messagetype=tomessagetype,
                                                                    filename=filename_translated,
                                                                    reference=transform.unique('messagecounter'),
                                                                    statust=OK,
                                                                    divtext=tscript)    #make outmessage object
                        
                        #~ botsglobal.logger.debug(_(u'Mappingscript "%(tscript)s" translates messagetype "%(messagetype)s" to messagetype "%(tomessagetype)s".'),
                                                #~ {'tscript':tscript,'messagetype':inn_splitup.ta_info['messagetype'],'tomessagetype':out_translated.ta_info['messagetype']})
                        translationscript,scriptfilename = botslib.botsimport('mappings',inn_splitup.ta_info['editype'],tscript)    #import mappingscript
                        alt_from_previous_run = inn_splitup.ta_info['alt']      #needed to check for infinite loop
                        doalttranslation = botslib.runscript(translationscript,scriptfilename,'main',inn=inn_splitup,out=out_translated)
                        botsglobal.logger.debug(_(u'Mappingscript "%(tscript)s" finished.'),{'tscript':tscript})
                        
                        #manipulate for some attributes after mapping script
                        if 'topartner' not in out_translated.ta_info:    #out_translated does not contain values from run......
                            out_translated.ta_info['topartner'] = inn_splitup.ta_info['topartner']
                        if 'botskey' in inn_splitup.ta_info:
                            inn_splitup.ta_info['reference'] = inn_splitup.ta_info['botskey']
                        if 'botskey' in out_translated.ta_info:    #out_translated does not contain values from run......
                            out_translated.ta_info['reference'] = out_translated.ta_info['botskey']
                            
                        #check the value received from the mappingscript to determine what to do in this while-loop. Handling of chained trasnlations.
                        if doalttranslation is None:    
                            #translation(s) are done; handle out-message 
                            out_translated.writeall()   #write result of translation.
                            #make translated record (if all is OK)
                            translated_dict = inn_splitup.ta_info.copy()
                            translated_dict.update(messagedict)
                            translated_dict.update(out_translated.ta_info)
                            run.translated.append(translated_dict)
                            del out_translated
                            break   #break out of while loop
                        elif isinstance(doalttranslation,dict):
                            #some extended cases; a dict is returned that contains 'instructions' for some type of chained translations
                            if 'type' not in doalttranslation or 'alt' not in doalttranslation:
                                raise botslib.BotsError(_(u"Mappingscript returned '%(alt)s'. This dict should not have 'type' and 'alt'."),{'alt':doalttranslation})
                            if alt_from_previous_run == doalttranslation['alt']:
                                number_of_loops_with_same_alt += 1
                            else:
                                number_of_loops_with_same_alt = 0
                            if doalttranslation['type'] == u'out_as_inn':
                                #do chained translation: use the out-object as inn-object, new out-object
                                #use case: detected error in incoming file; use out-object to generate warning email
                                handle_out_message(out_translated,ta_translated)
                                inn_splitup = out_translated    #out-object is now inn-object
                                if isinstance(inn_splitup,outmessage.fixed):    #for fixed: strip all values in node
                                    inn_splitup.root.stripnode()
                                inn_splitup.ta_info['alt'] = doalttranslation['alt']   #get the alt-value for the next chained translation
                                if not 'frompartner' in inn_splitup.ta_info:
                                    inn_splitup.ta_info['frompartner'] = ''
                                if not 'topartner' in inn_splitup.ta_info:
                                    inn_splitup.ta_info['topartner'] = ''
                                inn_splitup.ta_info.pop('statust')
                            elif doalttranslation['type'] == u'no_check_on_infinite_loop':
                                #do chained translation: allow many loops wit hsame alt-value.
                                #mapping script will have to handle this correctly.
                                number_of_loops_with_same_alt = 0
                                handle_out_message(out_translated,ta_translated)
                                del out_translated
                                inn_splitup.ta_info['alt'] = doalttranslation['alt']   #get the alt-value for the next chained translation
                            else:   #there is nothing else
                                raise botslib.BotsError(_(u'Mappingscript returned dict with an unknown "type": "%(doalttranslation)s".'),{'doalttranslation':doalttranslation})
                        else:  #note: this includes alt '' (empty string)
                            if alt_from_previous_run == doalttranslation:
                                number_of_loops_with_same_alt += 1
                            else:
                                number_of_loops_with_same_alt = 0
                            #do normal chained translation: same inn-object, new out-object
                            out_translated.writeall()
                            del out_translated
                            inn_splitup.ta_info['alt'] = doalttranslation   #get the alt-value for the next chained translation
                        if number_of_loops_with_same_alt > 10:
                            raise botslib.BotsError(_(u'Mappingscript returns same alt value over and over again (infinite loop?). Alt: "%(doalttranslation)s".'),{'doalttranslation':doalttranslation})
                    #end of while-loop (trans**********************************************************************************
                #exceptions file_out-level: exception in mappingscript or writing of out-file
                except:
                    #2 modes: either every error leads to skipping of  whole infile (old  mode) or errors in mappingscript/outfile only affect that branche
                    txt = botslib.txtexc()
                    print txt
                    messagedict['error'] += txt.strip()
                else:
                    pass
                    #~ print 'succes'
        #exceptions file_in-level
        except botslib.GotoException:   #edi-file is OK, file is passed-through after parsing.
            #~ edifile.handleconfirm(ta_fromfile,error=False)
            #~ botsglobal.logger.debug(_(u'Parse & passthrough for input file "%(filename)s".'),row)
            txt = botslib.txtexc()
            print txt
        except:
            txt = botslib.txtexc()
            messagedict['error'] += txt.strip()
            #~ edifile.handleconfirm(ta_fromfile,error=True)
            #~ botsglobal.logger.debug(u'Error in translating input file "%(filename)s":\n%(msg)s',{'filename':row['filename'],'msg':txt})
        else:
            pass

            
def mergemessages(run):
    names_envelope_criteria = ('editype','messagetype','frompartner','topartner','testindicator','charset','contenttype','envelope','rsrv3')
    merge_yes = {}
    merge_no = []       #can be non-unique: as a list
    #walk over run.translated; sort by envelope-criteria in 2 dicts
    for translated in run.translated:
        filename = translated.get('filename')
        infilename = translated.get('infilename')
        nrmessages = translated.get('nrmessages')
        if translated.get('merge'):
            env_criteria = tuple(translated.get(field) for field in names_envelope_criteria)
            if env_criteria in merge_yes:
                merge_yes[env_criteria][0].append(filename)
                merge_yes[env_criteria][1].append(infilename)
                merge_yes[env_criteria][2] += nrmessages
            else:
                merge_yes[env_criteria] = [[filename],[infilename],nrmessages]
        else:
            ta_info = dict((field,translated.get(field)) for field in names_envelope_criteria)
            ta_info['nrmessages'] = nrmessages
            merge_no.append((ta_info, [filename],[infilename]))
    #envelope
    for env_criteria,rest_of_info in merge_yes.iteritems():
        ta_info = dict(zip(names_envelope_criteria,env_criteria))
        ta_info['filename'] = transform.unique('bots_file_name')   #create filename for enveloped message
        ta_info['nrmessages'] = rest_of_info[2]
        ta_info['infilename'] = rest_of_info[1]      #for reference: list of infilenames
        ta_info['error'] = ''
        try:
            envelope.envelope(ta_info,rest_of_info[0])
        except:
            txt = botslib.txtexc()
            ta_info['error'] = txt.strip()
        finally:
            run.outgoing.append(ta_info)
    for ta_info,filenames,infilenames in merge_no:
        ta_info['filename'] = transform.unique('bots_file_name')   #create filename for enveloped message
        ta_info['infilename'] = infilenames      #for reference: list of infilenames
        ta_info['error'] = ''
        try:
            envelope.envelope(ta_info,filenames)
        except:
            txt = botslib.txtexc()
            ta_info['error'] = txt.strip()
        finally:
            run.outgoing.append(ta_info)

     
def write_outgoing(run):
    outputdir = botslib.join(run.outpath)
    botslib.dirshouldbethere(outputdir)
    for outgoing in run.outgoing:
        if not outgoing['error']:
            try:
                unique_filename = filename_formatter(run.outfilename,outgoing)
                tofilepath = botslib.join(outputdir,unique_filename)
                fromfilepath = botslib.abspathdata(outgoing['filename'])
                shutil.move(fromfilepath,tofilepath)
            except:
                txt = botslib.txtexc()
                outgoing.update({'error':txt})
            else:
                outgoing.update({'outfilename':tofilepath})
                
def filename_formatter(filename_mask,ta_info):
    class infilestr(str):
        ''' class for the infile-string that handles the specific format-options'''
        def __format__(self, format_spec):
            if not format_spec:
                return unicode(self)
            name,ext = os.path.splitext(unicode(self))
            if format_spec == 'ext':
                if ext.startswith('.'):
                    ext = ext[1:]
                return ext 
            if format_spec == 'name':
                return name 
            raise botslib.CommunicationOutError(_(u'Error in format of "{filename}": unknown format: "%(format)s".'),
                                                {'format':format_spec})
    unique = unicode(botslib.unique('bots_outgoing_file_name'))   #create unique part for filename
    tofilename = filename_mask.replace('*',unique)           #filename_mask is filename in channel where '*' is replaced by idta
    if '{' in tofilename :
        if botsglobal.ini.getboolean('acceptance','runacceptancetest',False):
            datetime_object = datetime.datetime.strptime("2013-01-23 01:23:45", "%Y-%m-%d %H:%M:%S")
        else:
            datetime_object = datetime.datetime.now()
        infilename = infilestr(os.path.basename(ta_info['infilename'][0])) #there is always an infile!
        tofilename = tofilename.format(infile=infilename,datetime=datetime_object,**ta_info)
    return tofilename


def trace(run):
    for incoming in run.incoming:
        if incoming['error']:
            run.errorinrun += 1
            continue
        #note: no search in translated: errors are in run.incoming
        #search in outgoing for envelope errors
        if run.outgoing:    #if no translation is done
            found = False
            for outgoing in run.outgoing:
                if incoming['infilename'] in outgoing['infilename']:
                    found = True
                    if outgoing['error']:
                        incoming['error'] += outgoing['error']
            if not found:   #debug sanity check
                raise Exception('incoming not in outgoing!')
    #count enveloping errors; delete these (no valid output)
    newlist = []
    for outgoing in run.outgoing:
        if outgoing['error']:
            run.errorinrun += 1
        else:
            newlist.append(outgoing)
    run.outgoing = newlist
        
def dict2xml(d):
    ''' convert python dictionary to xml.
    '''
    def makenode(tag,content):
        node = ET.Element(tag)
        if not content:
            pass    #empty element
        elif isinstance(content, basestring):
            node.text = content
        elif isinstance(content, list):
            node.tag = tag + 's'    #change node tag
            for element in content:
                node.append(makenode(tag, element))
        elif isinstance(content, dict):
            for key,value in content.items():
                node.append(makenode(key, value))
        else: 
            node.text = repr(content)
        return node
    assert isinstance(d, dict) and len(d) == 1
    for key,value in d.items():
        node = makenode(key,value)
    botslib.indent_xml(node)
    return ET.tostring(node)

def filter(lijst,names):
    return [dict((k,v) for k,v in d.items() if k in names) for d in lijst]
    
def report(run):
    in_filter = ('infilename','error','editype','messagetype')
    out_filter = ('outfilename','editype','messagetype','frompartner','topartner','infilename','nrmessages')
    xml_string = dict2xml({'root':{'nr_errors':run.errorinrun,'incoming':filter(run.incoming,in_filter),'outgoing':filter(run.outgoing,out_filter)}})
    print xml_string

def cleanup(run):
    shutil.rmtree(data_storage,ignore_errors=True)

'''
experiment with translation rule:

l = [{'alt': u'', 'fromeditype': u'edifact', 'frommessagetype': u'ORDERSD96AUNEAN008', 'frompartner': None, 'topartner': None, 'toeditype': u'xml', 'tomessagetype': u'orders', 'tscript': u'orders_edifact2xml'},
    {'alt': u'', 'fromeditype': u'edifact', 'frommessagetype': u'ORDERSD96AUNEAN008', 'frompartner': None, 'topartner': None, 'toeditype': u'xml', 'tomessagetype': u'orders', 'tscript': u'orders_edifact2xml'},
    ]

#first step: selecting for fromeditype,frommessagetype (active)
first = True
for d in l:
    if d['fromeditype'] != fromeditype or d['frommessagetype'] != frommessagetype:
        continue
    if (d['alt']!='' and d['alt']!=alt) or (d['frompartner'] is not None and d['frompartner']!=frompartner) or (d['frompartner'] is not None and d['frompartner']!=frompartner):
        continue
        
    if first:
        bestchoice = d
        continue
    #old algoritm: sort by alt, frompartner,topartner; pick first one in sort order
    #alt: either '' or value as in message
    if d['alt']
    #best choice: alt,frompartner and topartner
    #d: alt, 
    #if 

'''


''' Base library for bots. Botslib should not import code from other Bots-modules.'''
import sys
import os
import time
import codecs
import traceback
import socket
import urlparse
import urllib
import platform
import collections
import django
try:
    import importlib
except:
    import bots_importlib as importlib
from django.utils.translation import ugettext as _
#Bots-modules (no code)
from botsconfig import *    #constants
import botsglobal           #globals

#**********************************************************/**
#**************getters/setters for some globals***********************/**
#**********************************************************/**
def setrouteid(routeid):
    botsglobal.routeid = routeid

def getrouteid():
    return botsglobal.routeid

#**********************************************************/**
#***************** class  Transaction *********************/**
#**********************************************************/**
class _Transaction(object):
    ''' abstract class for db-ta.
        This class is used for communication with db-ta.
    '''
    #filtering values for db handling to avoid unknown fields in db.
    filterlist = ('statust','status','divtext','parent','child','script','frompartner','topartner','fromchannel','tochannel','editype','messagetype','merge',
                'testindicator','reference','frommail','tomail','contenttype','errortext','filename','charset','alt','idroute','nrmessages','retransmit',
                'confirmasked','confirmed','confirmtype','confirmidta','envelope','botskey','cc','rsrv1','filesize','numberofresends','rsrv3')
    processlist = [0]  #stack for bots-processes. last one is the current process; starts with 1 element in list: root

    def update(self,**ta_info):
        ''' Updates db-ta with named-parameters/dict.
            Use a filter to update only valid fields in db-ta
        '''
        setstring = ','.join([key+'=%('+key+')s' for key in ta_info if key in self.filterlist])
        if not setstring:   #nothing to update
            return
        ta_info['selfid'] = self.idta
        changeq(u'''UPDATE ta
                    SET '''+setstring+ '''
                    WHERE idta=%(selfid)s''',
                    ta_info)

    def delete(self):
        '''Deletes current transaction '''
        changeq(u'''DELETE FROM ta
                    WHERE idta=%(idta)s''',
                    {'idta':self.idta})

    def deletechildren(self):
        self.deleteonlychildren_core(self.idta)
        
    def deleteonlychildren_core(self,idta):
        for row in query(u'''SELECT idta 
                            FROM ta
                            WHERE parent=%(idta)s''',
                            {'idta':idta}):
            self.deleteonlychildren_core(row['idta'])
            changeq(u'''DELETE FROM ta
                        WHERE idta=%(idta)s''',
                        {'idta':row['idta' ]})
        
    def syn(self,*ta_vars):
        '''access of attributes of transaction as ta.fromid, ta.filename etc'''
        varsstring = ','.join(ta_vars)
        for row in query(u'''SELECT ''' + varsstring + '''
                              FROM ta
                              WHERE idta=%(idta)s''',
                              {'idta':self.idta}):
            self.__dict__.update(dict(row))

    def synall(self):
        '''access of attributes of transaction as ta.fromid, ta.filename etc'''
        self.syn(*self.filterlist)

    def copyta(self,status,**ta_info):
        ''' copy old transaction, return new transaction.
            parameters for new transaction are in ta_info (new transaction is updated with these values).
        '''
        script = _Transaction.processlist[-1]
        newidta = insertta(u'''INSERT INTO ta (script,  status,     parent,frompartner,topartner,fromchannel,tochannel,editype,messagetype,alt,merge,testindicator,reference,frommail,tomail,charset,contenttype,filename,idroute,nrmessages,botskey,envelope,rsrv3)
                                SELECT %(script)s,%(newstatus)s,idta,frompartner,topartner,fromchannel,tochannel,editype,messagetype,alt,merge,testindicator,reference,frommail,tomail,charset,contenttype,filename,idroute,nrmessages,botskey,envelope,rsrv3
                                FROM ta
                                WHERE idta=%(idta)s''',
                                {'idta':self.idta,'script':script,'newstatus':status})
        newta = OldTransaction(newidta)
        newta.update(**ta_info)
        return newta


class OldTransaction(_Transaction):
    ''' Resurrect old transaction '''
    def __init__(self,idta):
        self.idta = idta


class NewTransaction(_Transaction):
    ''' Generate new transaction. '''
    def __init__(self,**ta_info):
        updatedict = dict((key,value) for key,value in ta_info.iteritems() if key in self.filterlist)     #filter ta_info
        updatedict['script'] = self.processlist[-1]
        namesstring = ','.join([key for key in updatedict])
        varsstring = ','.join(['%('+key+')s' for key in updatedict])
        self.idta = insertta(u'''INSERT INTO ta (''' + namesstring + ''')
                                 VALUES   (''' + varsstring + ''')''',
                                updatedict)


class NewProcess(NewTransaction):
    ''' Create a new process (which is very much like a transaction).
        Used in logging of processes. 
        Each process is placed on stack processlist
    '''
    def __init__(self,functionname=''):
        super(NewProcess,self).__init__(filename=functionname,status=PROCESS,idroute=getrouteid())
        self.processlist.append(self.idta)

    def update(self,**ta_info):
        ''' update process, delete from process-stack. '''
        super(NewProcess,self).update(**ta_info)
        self.processlist.pop()


#**********************************************************/**
#*************************Database***********************/**
#**********************************************************/**
def addinfocore(change,where,wherestring):
    ''' core function for add/changes information in db-ta's.
    '''
    wherestring = ' WHERE idta > %(rootidta)s AND ' + wherestring
    counter = 0 #count the number of dbta changed
    for row in query(u'''SELECT idta FROM ta '''+wherestring,where):
        counter += 1
        ta_from = OldTransaction(row['idta'])
        ta_from.copyta(**change)     #make new ta from ta_from, using parameters from change
        ta_from.update(statust=DONE)    #update 'old' ta
    return counter
    
def addinfo(change,where):
    ''' change ta's to new phase: ta's are copied to new ta.
        returns the number of db-ta that have been changed.
        change (dict): values to change.
        where (dict): selection.
    '''
    if 'rootidta' not in where:
        where['rootidta'] = botsglobal.currentrun.get_minta4query()
    if 'statust' not in where:  #by default: look only for statust is OK
        where['statust'] = OK
    if 'statust' not in change: #by default: new ta is OK
        change['statust'] = OK
    wherestring = ' AND '.join([key+'=%('+key+')s ' for key in where if key != 'rootidta'])   #wherestring; does not use rootidta
    return addinfocore(change=change,where=where,wherestring=wherestring)

def updateinfocore(change,where,wherestring=''):
    ''' update info in ta's.
        where (dict) selects ta's,
        change (dict) sets values;
    '''
    wherestring = ' WHERE idta > %(rootidta)s AND ' + wherestring
    #change-dict: discard empty values. Change keys: this is needed because same keys can be in where-dict
    change2 = [(key,value) for key,value in change.iteritems() if value]
    if not change2:
        return
    changestring = ','.join([key+'=%(change_'+key+')s' for key,value in change2])
    where.update([('change_'+key,value) for key,value in change2])
    return changeq(u'''UPDATE ta SET ''' + changestring + wherestring,where)

def updateinfo(change,where):
    ''' update ta's.
        returns the number of db-ta that have been changed.
        change (dict): values to change.
        where (dict): selection.
    '''
    if 'rootidta' not in where:
        where['rootidta'] = botsglobal.currentrun.get_minta4query()
    if 'statust' not in where:  #by default: look only for statust is OK
        where['statust'] = OK
    if 'statust' not in change: #by default: new ta is OK
        change['statust'] = OK
    wherestring = ' AND '.join([key+'=%('+key+')s ' for key in where if key != 'rootidta'])   #wherestring for copy & done
    return updateinfocore(change=change,where=where,wherestring=wherestring)

def changestatustinfo(change,where):
    return updateinfo({'statust':change},where)

def query(querystring,*args):
    ''' general query. yields rows from query '''
    cursor = botsglobal.db.cursor()
    cursor.execute(querystring,*args)
    results =  cursor.fetchall()
    cursor.close()
    for result in results:
        yield result

def changeq(querystring,*args):
    '''general inset/update. no return'''
    cursor = botsglobal.db.cursor()
    try:
        cursor.execute(querystring,*args)
    except:
        botsglobal.db.rollback()    #rollback is needed for postgreSQL as this is also used by user scripts (eg via persist)
        raise
    botsglobal.db.commit()
    terug = cursor.rowcount
    cursor.close()
    return terug

def insertta(querystring,*args):
    ''' insert ta
        from insert get back the idta; this is different with postgrSQL.
    '''
    cursor = botsglobal.db.cursor()
    cursor.execute(querystring,*args)
    newidta = cursor.lastrowid
    if not newidta:   #if botsglobal.settings.DATABASE_ENGINE ==
        cursor.execute('''SELECT lastval() as idta''')
        newidta = cursor.fetchone()['idta']
    botsglobal.db.commit()
    cursor.close()
    return newidta

def unique_runcounter(domain,updatewith=None):
    ''' as unique, but per run of bots-engine.
    '''
    domain += 'bots_1_8_4_9_6'  #avoid using/mixing other values in botsglobal
    domain = domain.encode('unicode-escape')
    nummer = getattr(botsglobal,domain,0)
    if updatewith is None:
        nummer += 1
        updatewith = nummer
        if updatewith > sys.maxint-2:
            updatewith = 0
    setattr(botsglobal,domain,updatewith)
    return nummer

def unique(domein,updatewith=None):
    ''' generate unique number within range domain. Uses db to keep track of last generated number.
        3 use cases:
        - in acceptance: use unique_runcounter
        - if updatewith is not None: return current number, update database with updatewith
        - if updatewith is None: return current number plus 1; update database with  current number plus 1
            if domain not used before, initialize with 1.
    '''
    if botsglobal.ini.getboolean('acceptance','runacceptancetest',False):
        return unique_runcounter(domein)
    else:
        cursor = botsglobal.db.cursor()
        try:
            cursor.execute(u'''SELECT nummer FROM uniek WHERE domein=%(domein)s''',{'domein':domein})
            nummer = cursor.fetchone()['nummer']
            if updatewith is None:
                nummer += 1
                updatewith = nummer
                if updatewith > sys.maxint-2:
                    updatewith = 0
            cursor.execute(u'''UPDATE uniek SET nummer=%(nummer)s WHERE domein=%(domein)s''',{'domein':domein,'nummer':updatewith})
        except TypeError: #if domein does not exist, cursor.fetchone returns None, so TypeError
            cursor.execute(u'''INSERT INTO uniek (domein,nummer) VALUES (%(domein)s,1)''',{'domein': domein})
            nummer = 1
        botsglobal.db.commit()
        cursor.close()
        return nummer

def checkunique(domein, receivednumber):
    ''' to check if received number is sequential: value is compare with new generated number.
        if domain not used before, initialize it . '1' is the first value expected.
    '''
    newnumber = unique(domein)
    if newnumber  == receivednumber:
        return True
    else:
        #received number is not OK. Reset counter in database to previous value.  
        if botsglobal.ini.getboolean('acceptance','runacceptancetest',False):
            return False     #TODO: set the unique_runcounter
        else:
            changeq(u'''UPDATE uniek SET nummer=%(nummer)s WHERE domein=%(domein)s''',{'domein':domein,'nummer':newnumber-1})
            return False

#**********************************************************/**
#*************************Logging, Error handling********************/**
#**********************************************************/**
def sendbotserrorreport(subject,reporttext):
    ''' Send an email in case of errors or problems with bots-engine.
        Email is send to MANAGERS in config/settings.py.
        Email parameters are in config/settings.py (EMAIL_HOST, etc).
    '''
    if botsglobal.ini.getboolean('settings','sendreportiferror',False) and not botsglobal.ini.getboolean('acceptance','runacceptancetest',False):
        from django.core.mail import mail_managers
        try:
            mail_managers(subject, reporttext)
        except Exception as msg:
            botsglobal.logger.warning(u'Error in sending error report: %(msg)s',{'msg':msg})

def sendbotsemail(partner,subject,reporttext):
    ''' Send a simple email message to any bots partner.
        Mail is sent to all To: and cc: addresses for the partner (but send_mail does not support cc).
        Email parameters are in config/settings.py (EMAIL_HOST, etc).
    '''
    from django.core.mail import send_mail
    for row in query('''SELECT mail,cc FROM partner WHERE idpartner=%(partner)s''',{'partner':partner}):
        if row['mail']:
            recipient_list = row['mail'].split(',') + row['cc'].split(',')
            try:
                send_mail(subject, reporttext, botsglobal.settings.SERVER_EMAIL, recipient_list)
            except Exception as msg:
                botsglobal.logger.warning(u'Error sending email: %(msg)s',{'msg':msg})

def log_session(func):
    ''' used as decorator.
        The decorated functions are logged as processes.
        Errors in these functions are caught and logged.
    '''
    def wrapper(*args,**argv):
        try:
            ta_process = NewProcess(func.__name__)
        except:
            botsglobal.logger.exception(u'System error - no new process made')
            raise
        try:
            terug = func(*args,**argv)
        except:
            txt = txtexc()
            botsglobal.logger.debug(u'Error in process: %(txt)s',{'txt':txt})
            ta_process.update(statust=ERROR,errortext=txt)
        else:
            ta_process.update(statust=DONE)
            return terug
    return wrapper

def txtexc(mention_exception_type=True):
    ''' Process last exception, get an errortext.
        Errortext should be valid unicode.
    '''
    if botsglobal.ini and botsglobal.ini.getboolean('settings','debug',False):
        return safe_unicode(traceback.format_exc(limit=None))
    else:
        terug = safe_unicode(traceback.format_exc(limit=0))
        terug = terug.replace(u'Traceback (most recent call last):\n',u'')
        if not mention_exception_type:
            terug = terug.partition(': ')[2]
        return terug

def safe_unicode(value):
    ''' For errors: return best possible unicode...should never lead to errors.
    '''
    try:
        if isinstance(value, unicode):      #is already unicode, just return
            return value            
        elif isinstance(value, str):        #string/bytecode, encoding unknown.   
            for charset in ['utf_8','latin_1']:
                try:
                    return value.decode(charset, 'strict')  #decode strict
                except:
                    continue
            return value.decode('utf_8', 'ignore')  #decode as if it is utf-8, ignore errors.
        else:
            return unicode(value)
    except:
        try:
            return unicode(repr(value))
        except:
            return u'Error while displaying error'

class ErrorProcess(NewTransaction):
    ''' Used in logging of errors in processes: communication.py to indicate errors in receiving files (files have not been received)
    '''
    def __init__(self,functionname='',errortext='',channeldict=None):
        fromchannel = tochannel = ''
        if channeldict:
            if channeldict['inorout'] == 'in':
                fromchannel = channeldict['idchannel']
            else:
                tochannel = channeldict['idchannel']
        super(ErrorProcess,self).__init__(filename=functionname,status=PROCESS,idroute=getrouteid(),statust=ERROR,errortext=errortext,fromchannel=fromchannel,tochannel=tochannel)

#**********************************************************/**
#*************************import ***********************/**
#**********************************************************/**
def botsbaseimport(modulename):
    ''' Do a dynamic import.
        Errors/exceptions are handled in calling functions.
    '''
    return importlib.import_module(modulename.encode(sys.getfilesystemencoding()))

def botsimport(*args):
    ''' import modules from usersys.
        return: imported module, filename imported module;
        if not found or error in module: raise
    '''
    modulepath = '.'.join((botsglobal.usersysimportpath,) + args)             #assemble import string
    modulefile = join(botsglobal.ini.get('directories','usersysabs'),*args)   #assemble abs filename for errortexts; note that 'join' is function in this script-file.
    if modulepath in botsglobal.not_import:     #check if previous import failed (no need to try again).This eliminates eg lots of partner specific imports.
        botsglobal.logger.debug(_(u'No import of module "%(modulefile)s".'),{'modulefile':modulefile})
        raise BotsImportError(_(u'No import of module "%(modulefile)s".'),{'modulefile':modulefile})
    try:
        module = botsbaseimport(modulepath)
    except ImportError as msg:
        botsglobal.not_import.add(modulepath)
        botsglobal.logger.debug(_(u'No import of module "%(modulefile)s": %(txt)s.'),{'modulefile':modulefile,'txt':msg})
        raise BotsImportError(_(u'No import of module "%(modulefile)s": %(txt)s'),{'modulefile':modulefile,'txt':msg})
    except Exception as msg:
        botsglobal.logger.debug(_(u'Error in import of module "%(modulefile)s": %(txt)s.'),{'modulefile':modulefile,'txt':msg})
        raise ScriptImportError(_(u'Error in import of module "%(modulefile)s":\n%(txt)s'),{'modulefile':modulefile,'txt':msg})
    else:
        botsglobal.logger.debug(u'Imported "%(modulefile)s".',{'modulefile':modulefile})
        return module,modulefile
#**********************************************************/**
#*************************File handling os.path etc***********************/**
#**********************************************************/**
def join(*paths):
    '''Does does more as join.....
        - join the paths (compare os.path.join)
        - if path is not absolute, interpretate this as relative from bots directory.
        - normalize'''
    return os.path.normpath(os.path.join(botsglobal.ini.get('directories','botspath'),*paths))

def dirshouldbethere(path):
    if path and not os.path.exists(path):
        os.makedirs(path)
        return True
    return False

def abspath(soort,filename):
    ''' get absolute path for internal files; path is a section in bots.ini '''
    directory = botsglobal.ini.get('directories',soort)
    return join(directory,filename)

def abspathdata(filename):
    ''' abspathdata if filename incl dir: return absolute path; else (only filename): return absolute path (datadir)'''
    if '/' in filename: #if filename already contains path
        return join(filename)
    else:
        directory = botsglobal.ini.get('directories','data')
        datasubdir = filename[:-3]
        if not datasubdir:
            datasubdir = '0'
        return join(directory,datasubdir,filename)

def deldata(filename):
    ''' delete internal data file.'''
    filename = abspathdata(filename)
    try:
        os.remove(filename)
    except:
        #~ print 'not deleted', filename
        pass

def opendata(filename,mode,charset=None,errors='strict'):
    ''' open internal data file. if no encoding specified: read file raw/binary.'''
    filename = abspathdata(filename)
    if 'w' in mode:
        dirshouldbethere(os.path.dirname(filename))
    if charset:
        return codecs.open(filename,mode,charset,errors)
    else:
        return open(filename,mode)

def readdata(filename,charset=None,errors='strict'):
    ''' read internal data file in memory using the right encoding or no encoding'''
    filehandler = opendata(filename,'rb',charset,errors)
    content = filehandler.read()
    filehandler.close()
    return content

#**********************************************************/**
#*************************calling modules, programs***********************/**
#**********************************************************/**
def runscript(module,modulefile,functioninscript,**argv):
    ''' Execute userscript. Functioninscript is supposed to be there; if not AttributeError is raised.
        Often is checked in advance if Functioninscript does exist.
    '''
    botsglobal.logger.debug(u'Run userscript "%(functioninscript)s" in "%(modulefile)s".',
                            {'functioninscript':functioninscript,'modulefile':modulefile})
    functiontorun = getattr(module, functioninscript)
    try:
        return functiontorun(**argv)
    except:
        txt = txtexc()
        raise ScriptError(_(u'Userscript "%(modulefile)s": "%(txt)s".'),{'modulefile':modulefile,'txt':txt})

def tryrunscript(module,modulefile,functioninscript,**argv):
    if module and hasattr(module,functioninscript):
        runscript(module,modulefile,functioninscript,**argv)
        return True
    return False

def runscriptyield(module,modulefile,functioninscript,**argv):
    botsglobal.logger.debug(u'Run userscript "%(functioninscript)s" in "%(modulefile)s".',
                            {'functioninscript':functioninscript,'modulefile':modulefile})
    functiontorun = getattr(module, functioninscript)
    try:
        for result in functiontorun(**argv):
            yield result
    except:
        txt = txtexc()
        raise ScriptError(_(u'Script file "%(modulefile)s": "%(txt)s".'),{'modulefile':modulefile,'txt':txt})

#**********************************************************/**
#*************** confirmrules *****************************/**
#**********************************************************/**
def prepare_confirmrules():
    ''' as confirmrules are often used, read these into memory. Reason: performance.
        additional notes:
        - there are only a few confirmrules (10 would be a lot I guess).
        - indexing is not helpfull for confirmrules, this means that each time the whole confirmrule-tabel is scanned.
        - as confirmrules are used for incoming and outgoing (x12, edifact, email) this will almost always lead to better performance. 
    '''
    for confirmdict in query(u'''SELECT confirmtype,ruletype,idroute,idchannel_id as idchannel,frompartner_id as frompartner,topartner_id as topartner,messagetype,negativerule
                        FROM confirmrule
                        WHERE active=%(active)s
                        ORDER BY negativerule ASC
                        ''',
                        {'active':True}):
        botsglobal.confirmrules.append(confirmdict)

def set_asked_confirmrules(routedict,rootidta):
    ''' set 'ask confirmation/acknowledgements for x12 and edifact
    '''
    if not globalcheckconfirmrules(u'ask-x12-997') and not globalcheckconfirmrules(u'ask-edifact-CONTRL'):
        return
    for row in query('''SELECT parent,editype,messagetype,frompartner,topartner
                                FROM ta
                                WHERE idta>%(rootidta)s
                                AND status=%(status)s
                                AND statust=%(statust)s
                                AND (editype='edifact' OR editype='x12') ''',
                                {'status':FILEOUT,'statust':OK,'rootidta':rootidta}):
        if row['editype'] == 'x12':
            if row['messagetype'][:3] in ['997','999']:
                continue
            confirmtype = u'ask-x12-997'
        else:
            if row['messagetype'][:6] in ['CONTRL','APERAK']:
                continue
            confirmtype = u'ask-edifact-CONTRL'
        if not checkconfirmrules(confirmtype,idroute=routedict['idroute'],idchannel=routedict['tochannel'],
                                    topartner=row['topartner'],frompartner=row['frompartner'],messagetype=row['messagetype']):
            continue
        changeq('''UPDATE ta
                   SET confirmasked=%(confirmasked)s, confirmtype=%(confirmtype)s
                   WHERE idta=%(parent)s ''',
                   {'parent':row['parent'],'confirmasked':True,'confirmtype':confirmtype})

def globalcheckconfirmrules(confirmtype):
    ''' global check if confirmrules with this confirmtype is uberhaupt used. 
    '''
    for confirmdict in botsglobal.confirmrules:
        if confirmdict['confirmtype'] == confirmtype:
            return True
    return False

def checkconfirmrules(confirmtype,**kwargs):
    confirm = False       #boolean to return: confirm of not?
    #confirmrules are evaluated one by one; first the positive rules, than the negative rules.
    #this make it possible to include first, than exclude. Eg: send for 'all', than exclude certain partners.
    for confirmdict in botsglobal.confirmrules:
        if confirmdict['confirmtype'] != confirmtype:
            continue
        if confirmdict['ruletype'] == 'all':
            confirm = not confirmdict['negativerule']
        elif confirmdict['ruletype'] == 'route':
            if 'idroute' in kwargs and confirmdict['idroute'] == kwargs['idroute']:
                confirm = not confirmdict['negativerule']
        elif confirmdict['ruletype'] == 'channel':
            if 'idchannel' in kwargs and confirmdict['idchannel'] == kwargs['idchannel']:
                confirm = not confirmdict['negativerule']
        elif confirmdict['ruletype'] == 'frompartner':
            if 'frompartner' in kwargs and confirmdict['frompartner'] == kwargs['frompartner']:
                confirm = not confirmdict['negativerule']
        elif confirmdict['ruletype'] == 'topartner':
            if 'topartner' in kwargs and confirmdict['topartner'] == kwargs['topartner']:
                confirm = not confirmdict['negativerule']
        elif confirmdict['ruletype'] == 'messagetype':
            if 'messagetype' in kwargs and confirmdict['messagetype'] == kwargs['messagetype']:
                confirm = not confirmdict['negativerule']
        #~ print '>>>>>>>>>>>>', confirm,confirmtype,kwargs,confirmdict
    return confirm

#**********************************************************/**
#***************###############  misc.   #############
#**********************************************************/**
def set_database_lock():
    try:
        changeq(u'''INSERT INTO mutex (mutexk) VALUES (1)''')
    except:
        return False
    return True

def remove_database_lock():
    changeq('''DELETE FROM mutex WHERE mutexk=1''')

def check_if_other_engine_is_running():
    ''' bots-engine always connects to 127.0.0.1 port 28081 (or port as set in bots.ini).
        this  is a good way of detecting that another bots-engien is still running.
        problem is avoided anyway if using jobqueueserver.
    ''' 
    try:
        engine_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        port = botsglobal.ini.getint('settings','port',28081)
        engine_socket.bind(('127.0.0.1', port))
    except socket.error:
        engine_socket.close()
        raise
    else:
        return engine_socket

def trace_origin(ta,where=None):
    ''' bots traces back all from the current step/ta.
        where is a dict that is used to indicate a condition.
        eg:  {'status':EXTERNIN}
        If bots finds a ta for which this is true, the ta is added to a list.
        The list is returned when all tracing is done, and contains all ta's for which 'where' is True
    '''
    def trace_recurse(ta):
        ''' recursive
            walk over ta's backward (to origin).
            if condition is met, add the ta to a list
        '''
        for idta in get_parent(ta):
            donelijst.append(idta)
            taparent = OldTransaction(idta=idta)
            taparent.synall()
            for key,value in where.iteritems():
                if getattr(taparent,key) != value:
                    break
            else:   #all where-criteria are true; 
                teruglijst.append(taparent)
            trace_recurse(taparent)
    def get_parent(ta):
        ''' yields the parents of a ta '''
        if ta.parent:   #parent via the normal parent-attribute
            if ta.parent not in donelijst:
                yield ta.parent
        else:           #no parent via parent-link, so look via child-link
            for row in query('''SELECT idta
                                 FROM ta
                                 WHERE idta>%(minidta)s
                                 AND idta<%(maxidta)s
                                 AND child=%(idta)s''',
                                {'idta':ta.idta,'minidta':ta.script,'maxidta':ta.idta}):
                if row['idta'] in donelijst:
                    continue
                yield row['idta']

    donelijst = []
    teruglijst = []
    ta.synall()
    trace_recurse(ta)
    return teruglijst

def countoutfiles(idchannel,rootidta):
    ''' counts the number of edifiles to be transmitted via outchannel.'''
    for row in query('''SELECT COUNT(*) as count
                        FROM ta
                        WHERE idta>%(rootidta)s
                        AND status=%(status)s
                        AND statust=%(statust)s
                        AND tochannel=%(tochannel)s
                        ''',
                        {'status':FILEOUT,'statust':OK,'tochannel':idchannel,'rootidta':rootidta}):
        return row['count']

def lookup_translation(frommessagetype,fromeditype,alt,frompartner,topartner):
    ''' lookup the translation: frommessagetype,fromeditype,alt,frompartner,topartner -> mappingscript, tomessagetype, toeditype
    '''
    for row2 in query(u'''SELECT tscript,tomessagetype,toeditype
                            FROM translate
                            WHERE frommessagetype = %(frommessagetype)s
                            AND fromeditype = %(fromeditype)s
                            AND active=%(booll)s
                            AND (alt='' OR alt=%(alt)s)
                            AND (frompartner_id IS NULL OR frompartner_id=%(frompartner)s OR frompartner_id in (SELECT to_partner_id
                                                                                                                    FROM partnergroup
                                                                                                                    WHERE from_partner_id=%(frompartner)s ))
                            AND (topartner_id IS NULL OR topartner_id=%(topartner)s OR topartner_id in (SELECT to_partner_id
                                                                                                            FROM partnergroup
                                                                                                            WHERE from_partner_id=%(topartner)s ))
                            ORDER BY alt DESC,
                                     CASE WHEN frompartner_id IS NULL THEN 1 ELSE 0 END, frompartner_id ,
                                     CASE WHEN topartner_id IS NULL THEN 1 ELSE 0 END, topartner_id ''',
                            {'frommessagetype':frommessagetype,
                             'fromeditype':fromeditype,
                             'alt':alt,
                             'frompartner':frompartner,
                             'topartner':topartner,
                            'booll':True}):
        return row2['tscript'],row2['toeditype'],row2['tomessagetype']
        #translation is found; only the first one is used - this is what the ORDER BY in the query takes care of
    else:       #no translation found in translate table
        return None,None,None

def botsinfo():
    return [
            (_(u'served at port'),botsglobal.ini.getint('webserver','port',8080)),
            (_(u'platform'),platform.platform()),
            (_(u'machine'),platform.machine()),
            (_(u'python version'),platform.python_version()),
            (_(u'django version'),django.VERSION),
            (_(u'bots version'),botsglobal.version),
            (_(u'bots installation path'),botsglobal.ini.get('directories','botspath')),
            (_(u'config path'),botsglobal.ini.get('directories','config')),
            (_(u'botssys path'),botsglobal.ini.get('directories','botssys')),
            (_(u'usersys path'),botsglobal.ini.get('directories','usersysabs')),
            (u'DATABASE_ENGINE',botsglobal.settings.DATABASES['default']['ENGINE']),
            (u'DATABASE_NAME',botsglobal.settings.DATABASES['default']['NAME']),
            (u'DATABASE_USER',botsglobal.settings.DATABASES['default']['USER']),
            (u'DATABASE_HOST',botsglobal.settings.DATABASES['default']['HOST']),
            (u'DATABASE_PORT',botsglobal.settings.DATABASES['default']['PORT']),
            (u'DATABASE_OPTIONS',botsglobal.settings.DATABASES['default']['OPTIONS']),
            ]

def strftime(timeformat):
    if botsglobal.ini.getboolean('acceptance','runacceptancetest',False):
        return time.strftime(timeformat,time.strptime("2013-01-23 01:23:45", "%Y-%m-%d %H:%M:%S"))    #if acceptance test use fixed date/time
    else:
        return time.strftime(timeformat)

def settimeout(milliseconds):
    socket.setdefaulttimeout(milliseconds)    #set a time-out for TCP-IP connections

def updateunlessset(updatedict,fromdict):
    updatedict.update((key,value) for key, value in fromdict.iteritems() if key not in updatedict) #!!TODO when is this valid? Note: prevents setting charset from gramamr 
    #~ updatedict.update((key,value) for key, value in fromdict.iteritems() if key not in updatedict or not updatedict[key]) #!!TODO when is this valid? Note: prevents setting charset from gramamr 

def rreplace(org,old,new='',count=1):
    ''' string handling:
        replace old with new in org, max count times.
        with default values: remove last occurence of old in org.
    '''
    lijst = org.rsplit(old,count)
    return new.join(lijst)

def get_relevant_text_for_UnicodeError(msg):
    ''' see python doc for details of UnicodeError'''
    start = msg.start - 10 if msg.start >= 10 else 0
    return msg.object[start:msg.end+35]

def indent_xml(node, level=0,indentstring='    '):
    text2indent = "\n" + level*indentstring
    if len(node):
        if not node.text or not node.text.strip():
            node.text = text2indent + indentstring
        for subnode in node:
            indent_xml(subnode, level+1)
            if not subnode.tail or not subnode.tail.strip():
                subnode.tail = text2indent + indentstring
        if not subnode.tail or not subnode.tail.strip():
            subnode.tail = text2indent
    else:
        if level and (not node.tail or not node.tail.strip()):
            node.tail = text2indent

    
class Uri(object):
    ''' generate uri from parts/components
        - different forms of uri (eg with/without password)
        - general layout like 'scheme://user:pass@hostname:80/path/filename?query=argument#fragment'
        - checks: 1. what is required; 2. all parameters need to be valid
        Notes:
        - no filename: path ends with '/'
        Usage: uri = Uri(scheme='http',username='hje',password='password',hostname='test.com',port='80', path='')
        Usage: uri = Uri(scheme='http',hostname='test.com',port='80', path='test')
    '''
    def __init__(self, **kw):
        self._uri = dict(scheme=u'',username=u'',password=u'',hostname=u'',port=u'', path=u'', filename=u'',query={},fragment=u'')
        self.update(**kw)
    def update( self, **kw):
        self._uri.update(**kw)
    def uri(self, **kw):
        self.update(**kw)
        return self.__str__()
    def __str__(self):
        scheme   = self._uri['scheme'] + u':' if self._uri['scheme'] else u''
        password = u':' + self._uri['password'] if self._uri['password'] else u''
        userinfo = self._uri['username'] + password + u'@' if self._uri['username'] else u''
        port     = u':' + unicode(self._uri['port']) if self._uri['port'] else u''
        fullhost = self._uri['hostname'] + port if self._uri['hostname'] else u''
        authority = u'//' + userinfo + fullhost if fullhost else u''
        if self._uri['path'] or self._uri['filename']:
            terug = os.path.join(authority,self._uri['path'],self._uri['filename'])
        else:
            terug = authority
        return scheme + terug
#**********************************************************/**
#**************  Exception classes ***************************
#**********************************************************/**
class BotsError(Exception):
    ''' formats the error messages. Under all circumstances: give (reasonable) output, no errors.
        input (msg,*args,**kwargs) can be anything: strings (any charset), unicode, objects. Note that these are errors, so input can be 'not valid'!
        to avoid the risk of 'errors during errors' catch-all solutions are used.
        2 ways to raise Exceptions:
        - BotsError('tekst %(var1)s %(var2)s',{'var1':'value1','var2':'value2'})  ###this one is preferred!!
        - BotsError('tekst %(var1)s %(var2)s',var1='value1',var2='value2')
    '''
    def __init__(self, msg,*args,**kwargs):
        self.msg = safe_unicode(msg)
        if args:    ##expect args[0] to be a dict
            if isinstance(args[0],dict):
                xxx = args[0]
            else:
                xxx = {}
        else:
            xxx = kwargs
        self.xxx = collections.defaultdict(unicode)     #catch-all if var in string is not there
        for key,value in xxx.iteritems():
            self.xxx[safe_unicode(key)] = safe_unicode(value)
    def __unicode__(self):
        try:
            return self.msg%self.xxx    #this is already unicode
        except:
            return self.msg             #errors in self.msg; non supported format codes. Don't think this happen...
    def __str__(self):
        try:
            return (self.msg%self.xxx).encode('utf-8','ignore')
        except:
            return self.msg.encode('utf-8','ignore')            #errors in self.msg; non supported format codes. Don't think this happen...
            
class CodeConversionError(BotsError):
    pass
class CommunicationError(BotsError):
    pass
class CommunicationInError(BotsError):
    pass
class CommunicationOutError(BotsError):
    pass
class EanError(BotsError):
    pass
class GrammarError(BotsError):
    pass
class InMessageError(BotsError):
    pass
class LockedFileError(BotsError):
    pass
class MessageError(BotsError):
    pass
class MessageRootError(BotsError):
    pass
class MappingRootError(BotsError):
    pass
class MappingFormatError(BotsError):
    pass
class OutMessageError(BotsError):
    pass
class PanicError(BotsError):
    pass
class PersistError(BotsError):
    pass
class PluginError(BotsError):
    pass
class BotsImportError(BotsError):    #import script or recursivly imported scripts not there
    pass
class ScriptImportError(BotsError):    #import errors in userscript; userscript is there
    pass
class ScriptError(BotsError):          #runtime errors in a userscript
    pass
class TraceError(BotsError):
    pass
class TranslationNotFoundError(BotsError):
    pass
class GotoException(BotsError):     #sometimes it is simplest to raise an error, and catch it rightaway. Like a goto ;-)  
    pass
class FileTooLargeError(BotsError):
    pass



from django.utils.translation import ugettext as _
#bots-modules
import botslib
import node
import botsglobal
import grammar
from botsconfig import *


class Message(object):
    ''' abstract class; represents a edi message.
        is subclassed as outmessage or inmessage object.
    '''
    def __init__(self,ta_info):
        self.ta_info = ta_info      #here ta_info is only filled with parameters from db-ta
        self.errorlist = []         #collect non-fatal errors in the edi file; used in reporting errors.
        self.errorfatal = True     #store fatal errors: errors that stop the processing of the file
        self.messagetypetxt = ''    #used in reporting errors.
        self.messagecount = 0       #count messages in edi file; used in reporting errors.

    def add2errorlist(self,errortxt):
        ''' Handle non-fatal parse errors.
        '''
        #~ raise botslib.MessageError(u'For unit format test')     #UNITTEST_CORRECTION 
        if len(self.errorlist) < botsglobal.ini.getint('settings','max_number_errors',10):
            self.errorlist.append(self.messagetypetxt + errortxt)
        elif len(self.errorlist) == botsglobal.ini.getint('settings','max_number_errors',10):
            self.errorlist.append((_(u'Found at least %(max_number_errors)s errors.')%
                                        {'max_number_errors':len(self.errorlist)}))
        else:
            #more than max_number_errors: stop adding new errors to list.
            pass

    def checkforerrorlist(self):
        ''' examine the message-object for errors; 
        '''
        if self.errorfatal:     #for fatal errors: (try to) get information like partners for edi file
            self.try_to_retrieve_info()
        if self.errorlist:
            raise botslib.MessageError(u''.join(self.errorlist))

    def try_to_retrieve_info(self):
        ''' when edi-file is not correct, (try to) get info about eg partnerID's in message
            method is specified in subclasses.
        ''' 
        pass
        
    def messagegrammarread(self,typeofgrammarfile='grammars'):
        ''' read grammar for a message/envelope.
        '''
        #read grammar for message.
        #starts with default values from grammar.py; values are overruled by envelope settings; values are overrules by messagetype setting.
        self.defmessage = grammar.grammarread(self.ta_info['editype'],self.ta_info['messagetype'],typeofgrammarfile)
        #write values from grammar to self.ta_info - unless these values are already set (eg by mappingscript)
        botslib.updateunlessset(self.ta_info,self.defmessage.syntax)
        if not self.ta_info['charset']:
            self.ta_info['charset'] = self.defmessage.syntax['charset']      #always use charset of edi file.

    @staticmethod
    def display(lex_records):
        ''' for debugging: display lexed records.'''
        for lex_record in lex_records:
            counter = 0
            for veld in lex_record:
                if counter == 0:
                    print '%s    (Record-id)'%(veld[VALUE])
                else:
                    if veld[SFIELD] == 0:
                        print '    %s    (field)'%(veld[VALUE])
                    elif veld[SFIELD] == 1:
                        print '        %s    (sub)'%(veld[VALUE])
                    elif veld[SFIELD] == 2:
                        print '        %s    (rep)'%(veld[VALUE])
                    else:
                        print '    %s    (???)'%(veld[VALUE])
                counter += 1

    @staticmethod
    def mpathformat(mpath):
        ''' mpath is eg: ['UNH', 'NAD'], formatted is: 'UNH-NAD'.
        '''
        return '-'.join(mpath)
        
    def checkmessage(self,node_instance,defmessage,subtranslation=False):
        ''' The node tree is check, sorted, fields are formatted etc against grammar. (so far only minimal tests have been done during processing)
            For checking translation & subtranslation
            parameter 'subtranslation' only used for reporting
            some different cases:
            - empy root.record, root.children filled:
            - edifact, x12, tradacoms: each child is an envelope. Check each envelope. (use mailbag to have one UNB per node-tree here)
            - csv nobotsid: each child is a record. Check all records in one check
            - xml, json:
            root.record filled, root.children filled: outgoing messages.
        '''
        if not self.ta_info['has_structure']:
            return
        if node_instance.record:        #root record contains information; write whole tree in one time
            self._checkonemessage(node_instance,defmessage,subtranslation)
        else:
            for childnode in node_instance.children:
                self._checkonemessage(childnode,defmessage,subtranslation)

    def _checkonemessage(self,node_instance,defmessage,subtranslation):
        structure = defmessage.structure
        if node_instance.record['BOTSID'] != structure[0][ID]:
            raise botslib.MessageRootError(_(u'[G50]: Grammar "%(grammar)s" starts with record "%(grammarroot)s"; but in edi-file found start-record "%(root)s".'),
                                        {'root':node_instance.record['BOTSID'],'grammarroot':structure[0][ID],'grammar':defmessage.grammarname})
        self._checkifrecordsingrammar(node_instance,structure[0],defmessage.grammarname)
        self._canonicaltree(node_instance,structure[0])
        if not subtranslation and botsglobal.ini.getboolean('settings','readrecorddebug',False):       #should the content of the message (the records read) be logged.
            self._logmessagecontent(node_instance)

    def _checkifrecordsingrammar(self,node_instance,structure,grammarname):
        ''' check for every node if in grammar
            recursive
        '''
        deletelist = []       #list of records not in the grammar; these records are deleted at end of function
        self._checkiffieldsingrammar(node_instance,structure)     #check if fields are known in grammar
        if 'messagetype' in node_instance.queries:   #determine if SUBTRANSLATION starts; do not check (is already checked)
            return
        if node_instance.children and LEVEL not in structure:            #if record has children, but these are not in the grammar
            if self.ta_info['checkunknownentities']:
                self.add2errorlist(_(u'[S01]%(linpos)s: Record "%(record)s" in message has children, but these are not in grammar "%(grammar)s". Found record "%(xx)s".\n')%
                                    {'linpos':node_instance.linpos(),'record':node_instance.record['BOTSID'],'grammar':grammarname,'xx':node_instance.children[0].record['BOTSID']})
            node_instance.children = []
            return
        for childnode in node_instance.children:          #for every record/childnode:
            for record_definition in structure[LEVEL]:                   #search in grammar-records
                if childnode.record['BOTSID'] == record_definition[ID]:
                    #found record in grammar
                    #check recursive:
                    self._checkifrecordsingrammar(childnode,record_definition,grammarname)
                    break    #record/childnode is in gramar; go to check next record/childnode
            else:   #record/childnode in not in grammar
                if self.ta_info['checkunknownentities']:
                    self.add2errorlist(_(u'[S02]%(linpos)s: Unknown record "%(record)s" in message.\n')%
                                        {'linpos':node_instance.linpos(),'record':childnode.record['BOTSID']})
                deletelist.append(childnode)
        for child in deletelist:
            node_instance.children.remove(child)


    def _checkiffieldsingrammar(self,node_instance,record_definition):
        ''' checks for every field in record if field exists in record_definition (from grammar).
            for inmessage of type (var,fixed,??) this is not needed 
        '''
        for field in node_instance.record.keys():     #check every field in the record
            if field == 'BOTSIDnr':     #BOTSIDnr is not in grammar, so skip check
                continue
            for field_definition in record_definition[FIELDS]:
                if field_definition[ISFIELD]:    #if field (no composite)
                    if field == field_definition[ID]:
                        break   #OK!
                else:   #if composite
                    if field_definition[MAXREPEAT] == 1:    #non-repeating composite 
                        for grammarsubfield in field_definition[SUBFIELDS]:   #loop subfields
                            if field == grammarsubfield[ID]:
                                break   #break out of grammarsubfield-for-loop ->goto break out of field_definition-for-loop
                        else:
                            continue    #nothing found; continue with next gammarfield
                        break   #break out of field_definition-for-loop
                    else: #repeating composite
                        if field == field_definition[ID]:
                            #OK. Contents is a list of dicts;
                            #TODO: check for each dict if sub-fields exist in grammar. 
                            break
            else:           #field not found in grammar
                if self.ta_info['checkunknownentities']:
                    self.add2errorlist(_(u'[F01]%(linpos)s: Record: "%(mpath)s" has unknown field "%(field)s".\n')%
                                            {'linpos':node_instance.linpos(),'field':field,'mpath':self.mpathformat(record_definition[MPATH])})
                del node_instance.record[field]

    def _canonicaltree(self,node_instance,structure):
        ''' For nodes: check min and max occurence; sort the records conform grammar
        '''
        sortednodelist = []
        self._canonicalfields(node_instance,structure)    #handle fields of this record
        if node_instance.structure is None:
            node_instance.structure = structure
        if LEVEL in structure:
            for record_definition in structure[LEVEL]:  #for every record_definition (in grammar) of this level
                count = 0                           #count number of occurences of record
                for childnode in node_instance.children:            #for every node in mpathtree; SPEED: delete nodes from list when found
                    if childnode.record['BOTSID'] != record_definition[ID] or childnode.record['BOTSIDnr'] != record_definition[BOTSIDNR]:   #if it is not the right NODE":
                        continue
                    count += 1
                    self._canonicaltree(childnode,record_definition)         #use rest of index in deeper level
                    sortednodelist.append(childnode)
                if record_definition[MIN] > count:
                    self.add2errorlist(_(u'[S03]%(linpos)s: Record "%(mpath)s" occurs %(count)d times, min is %(mincount)d.\n')%
                                        {'linpos':node_instance.linpos(),'mpath':self.mpathformat(record_definition[MPATH]),'count':count,'mincount':record_definition[MIN]})
                if record_definition[MAX] < count:
                    self.add2errorlist(_(u'[S04]%(linpos)s: Record "%(mpath)s" occurs %(count)d times, max is %(maxcount)d.\n')%
                                        {'linpos':node_instance.linpos(),'mpath':self.mpathformat(record_definition[MPATH]),'count':count,'maxcount':record_definition[MAX]})
            node_instance.children = sortednodelist

    def _canonicalfields(self,node_instance,record_definition):
        ''' For all fields: check M/C, format.
            Fields are not sorted (a dict can not be sorted).
            Fields are never added.
        '''
        noderecord = node_instance.record
        for field_definition in record_definition[FIELDS]:       #loop over fields in grammar
            if field_definition[ISFIELD]:    #if field (no composite)
                if field_definition[MAXREPEAT] == 1:    #if non-repeating
                    value = noderecord.get(field_definition[ID])
                    if not value:
                        if field_definition[MANDATORY]:
                            self.add2errorlist(_(u'[F02]%(linpos)s: Record "%(mpath)s" field "%(field)s" is mandatory.\n')%
                                                {'linpos':node_instance.linpos(),'mpath':self.mpathformat(record_definition[MPATH]),'field':field_definition[ID]})
                        continue
                    noderecord[field_definition[ID]] = self._formatfield(value,field_definition,record_definition,node_instance)
                else: #repeating field;
                    #a list of values; values can be empty or None; at least one field should have value, else dropped
                    valuelist = noderecord.get(field_definition[ID])
                    if valuelist is None:   #empty lists are already catched in node.put()
                        if field_definition[MANDATORY]:
                            self.add2errorlist(_(u'[F41]%(linpos)s: Record "%(mpath)s" repeating field "%(field)s" is mandatory.\n')%
                                                {'linpos':node_instance.linpos(),'mpath':self.mpathformat(record_definition[MPATH]),'field':field_definition[ID]})
                        continue
                    if not isinstance(valuelist,list):
                        raise botslib.MappingFormatError(_(u'Repeating field: must be a list: put(%(valuelist)s)'),{'mpath':valuelist})
                    if len(valuelist) > field_definition[MAXREPEAT]:
                        self.add2errorlist(_(u'[F42]%(linpos)s: Record "%(mpath)s" repeating field "%(field)s" occurs %(occurs)s times, max is %(max)s.\n')%
                                            {'linpos':node_instance.linpos(),'mpath':self.mpathformat(record_definition[MPATH]),'field':field_definition[ID],
                                             'occurs':len(valuelist),'max':field_definition[MAXREPEAT]})
                    newlist = []
                    repeating_field_has_data = False
                    for value in valuelist:
                        if value is None:
                            value = ''
                        else:
                            value = unicode(value).strip()
                            if value:
                                repeating_field_has_data = True
                        newlist.append(self._formatfield(value,field_definition,record_definition,node_instance))
                    if not repeating_field_has_data:
                        if field_definition[MANDATORY]:
                            self.add2errorlist(_(u'[F43]%(linpos)s: Record "%(mpath)s" repeating field "%(field)s" is mandatory.\n')%
                                                {'linpos':node_instance.linpos(),'mpath':self.mpathformat(record_definition[MPATH]),'field':field_definition[ID]})
                        del noderecord[field_definition[ID]]
                        continue
                    noderecord[field_definition[ID]] = newlist
            else:                                   #composite
                if field_definition[MAXREPEAT] == 1:    #non-repeating compostie
                    #first check if there is any data att all in this composite
                    for grammarsubfield in field_definition[SUBFIELDS]:
                        if noderecord.get(grammarsubfield[ID]):
                            break   #composite has data.
                    else:           #composite has no data
                        if field_definition[MANDATORY]:
                            self.add2errorlist(_(u'[F03]%(linpos)s: Record "%(mpath)s" composite "%(field)s" is mandatory.\n')%
                                                {'linpos':node_instance.linpos(),'mpath':self.mpathformat(record_definition[MPATH]),'field':field_definition[ID]})
                        continue    #there is no data in composite, so do nothing
                    #there is data in the composite!
                    for grammarsubfield in field_definition[SUBFIELDS]:   #loop subfields
                        value = noderecord.get(grammarsubfield[ID])
                        if not value:
                            if grammarsubfield[MANDATORY]:
                                self.add2errorlist(_(u'[F04]%(linpos)s: Record "%(mpath)s" subfield "%(field)s" is mandatory.\n')%
                                                    {'linpos':node_instance.linpos(),'mpath':self.mpathformat(record_definition[MPATH]),'field':grammarsubfield[ID]})
                            continue
                        noderecord[grammarsubfield[ID]] = self._formatfield(value,grammarsubfield,record_definition,node_instance)
                else:   #if repeating composite: list of dicts
                    valuelist = noderecord.get(field_definition[ID])
                    #~ print 'valuelist',valuelist
                    if valuelist is None:   #empty lists are catched in node.put()
                        if field_definition[MANDATORY]:
                            self.add2errorlist(_(u'[F44]%(linpos)s: Record "%(mpath)s" repeating composite "%(field)s" is mandatory.\n')%
                                                {'linpos':node_instance.linpos(),'mpath':self.mpathformat(record_definition[MPATH]),'field':field_definition[ID]})
                        continue
                    if not isinstance(valuelist,list):
                        raise botslib.MappingFormatError(_(u'Repeating composite: must be a list: put(%(valuelist)s)'),{'mpath':valuelist})
                    if len(valuelist) > field_definition[MAXREPEAT]:
                        self.add2errorlist(_(u'[F45]%(linpos)s: Record "%(mpath)s" repeating composite "%(field)s" occurs %(occurs)s times, max is %(max)s.\n')%
                                            {'linpos':node_instance.linpos(),'mpath':self.mpathformat(record_definition[MPATH]),'field':field_definition[ID],
                                             'occurs':len(valuelist),'max':field_definition[MAXREPEAT]})
                    #is a list of composites; each composite is a dict.
                    #loop over composites/dicts, create a new list
                    #- dict can be empty: {}. check if: composite contains data or is empty
                    #- if composite is not-empty, check M/C of elements in composite
                    #- keys should be basestring.
                    #- convert values to unicode 
                    #- check if there is data for whole repeating composite
                    newlist = []
                    repeating_composite_has_data = False
                    for comp in valuelist:
                        if not isinstance(comp,dict):
                            raise botslib.MappingFormatError(_(u'Repeating composite: each composite must be a dict: put(%(valuelist)s)'),{'mpath':valuelist})
                        #check each dict, convert values to unicode
                        #also: check if dict has data at all
                        composite_has_data = False
                        for key,value in comp.iteritems():
                            if not isinstance(key,basestring):
                                raise botslib.MappingFormatError(_(u'Repeating composite: keys must be strings: put(%(mpath)s)'),{'mpath':valuelist})
                            if value is None:
                                comp[key] = u''
                            else:
                                comp[key] = unicode(value).strip()  #leading and trailing spaces are stripped from the values
                                if comp[key]:
                                    composite_has_data = True
                        if composite_has_data:
                            repeating_composite_has_data = True
                            for grammarsubfield in field_definition[SUBFIELDS]:   #loop subfields
                                value = comp.get(grammarsubfield[ID])
                                if not value:
                                    if grammarsubfield[MANDATORY]:
                                        self.add2errorlist(_(u'[F46]%(linpos)s: Record "%(mpath)s" subfield "%(field)s" in repeating composite is mandatory.\n')%
                                                            {'linpos':node_instance.linpos(),'mpath':self.mpathformat(record_definition[MPATH]),'field':grammarsubfield[ID]})
                                    continue
                                comp[grammarsubfield[ID]] = self._formatfield(value,grammarsubfield,record_definition,node_instance)
                        else:
                            comp = {}
                        newlist.append(comp)
                    if not repeating_composite_has_data: 
                        if field_definition[MANDATORY]: 
                            self.add2errorlist(_(u'[F47]%(linpos)s: Record "%(mpath)s" repeating composite "%(field)s" is mandatory.\n')%
                                            {'linpos':node_instance.linpos(),'mpath':self.mpathformat(record_definition[MPATH]),'field':field_definition[ID]})
                        
                        del noderecord[field_definition[ID]]
                    else:
                        noderecord[field_definition[ID]] = newlist


    def _logmessagecontent(self,node_instance):
        botsglobal.logger.debug(u'Record "%(BOTSID)s":',node_instance.record)
        self._logfieldcontent(node_instance.record)    #handle fields of this record
        for child in node_instance.children:
            self._logmessagecontent(child)

    @staticmethod
    def _logfieldcontent(noderecord):
        for key,value in noderecord.iteritems():
            if key not in ['BOTSID','BOTSIDnr']:
                botsglobal.logger.debug(u'    "%(key)s" : "%(value)s"',{'key':key,'value':value})

    #***************************************************************************
    #* methods below pass call to node.Node ************************************
    #***************************************************************************
    def getrecord(self,*mpaths):
        if self.root.record is None:
            raise botslib.MappingRootError(_(u'getrecord(%(mpath)s): "root" of incoming message is empty; either split messages or use inn.getloop'),
                                            {'mpath':mpaths})
        return self.root.getrecord(*mpaths)

    def change(self,where,change):
        ''' query tree (self.root) with where; if found replace with change; return True if change, return False if not changed.'''
        if self.root.record is None:
            raise botslib.MappingRootError(_(u'change(%(where)s,%(change)s): "root" of incoming message is empty; either split messages or use inn.getloop'),
                                            {'where':where,'change':change})
        return self.root.change(where,change)

    def delete(self,*mpaths):
        ''' query tree (self.root) with mpath; delete if found. return True if deleted, return False if not deleted.'''
        if self.root.record is None:
            raise botslib.MappingRootError(_(u'delete(%(mpath)s): "root" of incoming message is empty; either split messages or use inn.getloop'),
                                            {'mpath':mpaths})
        return self.root.delete(*mpaths)

    def get(self,*mpaths):
        ''' query tree (self.root) with mpath; get value (string); get None if not found.'''
        if self.root.record is None:
            raise botslib.MappingRootError(_(u'get(%(mpath)s): "root" of incoming message is empty; either split messages or use inn.getloop'),
                                            {'mpath':mpaths})
        return self.root.get(*mpaths)

    def getnozero(self,*mpaths):
        ''' like get, returns None is value is zero (0) or not numeric.
            Is sometimes usefull in mapping.'''
        if self.root.record is None:
            raise botslib.MappingRootError(_(u'getnozero(%(mpath)s): "root" of incoming message is empty; either split messages or use inn.getloop'),
                                            {'mpath':mpaths})
        return self.root.getnozero(*mpaths)

    def getdecimal(self,*mpaths):
        ''' like get, returns None is value is zero (0) or not numeric.
            Is sometimes usefull in mapping.'''
        if self.root.record is None:
            raise botslib.MappingRootError(_(u'getdecimal(%(mpath)s): "root" of incoming message is empty; either split messages or use inn.getloop'),
                                            {'mpath':mpaths})
        return self.root.getdecimal(*mpaths)

    def getcount(self):
        ''' count number of nodes in self.root. Number of nodes is number of records.'''
        return self.root.getcount()

    def getcountoccurrences(self,*mpaths):
        ''' count number of nodes in self.root. Number of nodes is number of records.'''
        return len(list(self.getloop(*mpaths)))

    def getcountsum(self,*mpaths):
        ''' return the sum for all values found in mpath. Eg total number of ordered quantities.'''
        if self.root.record is None:
            raise botslib.MappingRootError(_(u'get(%(mpath)s): "root" of incoming message is empty; either split messages or use inn.getloop'),
                                                {'mpath':mpaths})
        return self.root.getcountsum(*mpaths)

    def getloop(self,*mpaths):
        ''' query tree with mpath; generates all the nodes. Is typically used as: for record in inn.get(mpath):
        '''
        if self.root.record:    #self.root is a real root
            for terug in self.root.getloop(*mpaths): #search recursive for rest of mpaths
                yield terug
        else:   #self.root is dummy root
            for childnode in self.root.children:
                for terug in childnode.getloop(*mpaths): #search recursive for rest of mpaths
                    yield terug

    def put(self,*mpaths,**kwargs):
        if self.root.record is None and self.root.children:
            raise botslib.MappingRootError(_(u'put(%(mpath)s): "root" of outgoing message is empty; use out.putloop'),
                                            {'mpath':mpaths})
        return self.root.put(*mpaths,**kwargs)

    def putloop(self,*mpaths):
        if not self.root.record:    #no input yet, and start with a putloop(): dummy root
            if len(mpaths) == 1:
                self.root.append(node.Node(record=mpaths[0]))
                return self.root.children[-1]
            else:
                raise botslib.MappingRootError(_(u'putloop(%(mpath)s): mpath too long???'),
                                                {'mpath':mpaths})
        return self.root.putloop(*mpaths)

    def sort(self,*mpaths):
        if self.root.record is None:
            raise botslib.MappingRootError(_(u'get(%(mpath)s): "root" of message is empty; either split messages or use inn.getloop'),
                                            {'mpath':mpaths})
        self.root.sort(*mpaths)
